from django.db import models


class VisibilityModel(models.Model):
    is_public = models.BooleanField(
        default=False, help_text="If the entity is public or private."
    )

    class Meta:
        abstract = True



from haupt.common.events import event_actions, event_subjects
from haupt.common.events.event import ActorTopEntityEvent
from haupt.common.events.registry.attributes import PROJECT_OWNER_ATTRIBUTES

PROJECT_CREATED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.CREATED, event_subjects.ACTOR
)
PROJECT_UPDATED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.UPDATED, event_subjects.ACTOR
)
PROJECT_SETTINGS_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.SETTINGS, event_subjects.ACTOR
)
PROJECT_DELETED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.DELETED, event_subjects.ACTOR
)
PROJECT_VIEWED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.VIEWED, event_subjects.ACTOR
)
PROJECT_RUNS_VIEWED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.RUNS_VIEWED, event_subjects.ACTOR
)
PROJECT_STATS_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.STATS, event_subjects.ACTOR
)

EVENTS = {
    PROJECT_CREATED_ACTOR,
    PROJECT_UPDATED_ACTOR,
    PROJECT_SETTINGS_ACTOR,
    PROJECT_VIEWED_ACTOR,
    PROJECT_DELETED_ACTOR,
    PROJECT_RUNS_VIEWED_ACTOR,
    PROJECT_STATS_ACTOR,
}


class ProjectCreatedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_CREATED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectUpdatedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_UPDATED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectSettingsActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_SETTINGS_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectDeletedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_DELETED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectViewedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_VIEWED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectRunsViewedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_RUNS_VIEWED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectStatsActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_STATS_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES



from typing import Any, Dict, List, Optional

from clipped.utils.lists import to_list

from haupt.common import auditor
from haupt.common.events.registry.run import (
    RUN_DONE,
    RUN_FAILED,
    RUN_NEW_STATUS,
    RUN_RESUMED,
    RUN_SKIPPED,
    RUN_STOPPED,
    RUN_SUCCEEDED,
)
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from polyaxon.schemas import LifeCycle, V1StatusCondition, V1Statuses


def get_run_conditions(entity) -> List[V1StatusCondition]:
    if entity.status_conditions:
        status_conditions = to_list(entity.status_conditions, check_none=True)
    else:
        status_conditions = []
    return [V1StatusCondition.get_condition(**c) for c in status_conditions]


def sort_conditions(status_conditions):
    return sorted(
        status_conditions,
        key=lambda x: V1StatusCondition.get_last_update_time(
            x.get("last_transition_time")
        ),
    )


def set_entity_status(entity, condition: V1StatusCondition):
    entity.status = condition.type

    if condition:
        status_conditions = None
        if entity.status_conditions:
            status_conditions = to_list(entity.status_conditions, check_none=True)
            last_condition = V1StatusCondition.get_condition(**status_conditions[-1])
            if last_condition == condition:
                status_conditions[-1] = condition.to_dict()
            else:
                status_conditions.append(condition.to_dict())
        elif condition:
            status_conditions = [condition.to_dict()]
        if status_conditions:
            entity.status_conditions = status_conditions

    return entity


def new_status(
    entity,
    condition: V1StatusCondition,
    additional_fields: Optional[List[str]] = None,
    force: bool = False,
):
    previous_status = entity.status
    if condition.type == V1Statuses.CREATED:
        return previous_status
    if previous_status == V1Statuses.STOPPING and not LifeCycle.is_done(condition.type):
        return previous_status
    # Do not override final status
    if LifeCycle.is_done(previous_status) and not force:
        return previous_status

    entity = set_entity_status(entity=entity, condition=condition)

    LifeCycle.set_started_at(entity=entity)
    LifeCycle.set_finished_at(entity=entity)
    additional_fields = additional_fields or []
    entity.save(
        update_fields=additional_fields
        + [
            "status_conditions",
            "status",
            "started_at",
            "updated_at",
            "finished_at",
            "wait_time",
            "duration",
        ]
    )

    return previous_status


def bulk_new_entity_status(
    model_class,
    entities: List[Any],
    condition: V1StatusCondition,
    additional_fields: Optional[List[str]] = None,
):
    for entity in entities:
        set_entity_status(entity=entity, condition=condition)
    additional_fields = additional_fields or []
    model_class.all.bulk_update(
        entities, additional_fields + ["status_conditions", "status"]
    )


def bulk_new_run_status(
    runs: List[BaseRun],
    condition: V1StatusCondition,
    additional_fields: Optional[List[str]] = None,
):
    bulk_new_entity_status(
        model_class=Models.Run,
        entities=runs,
        condition=condition,
        additional_fields=additional_fields,
    )


def new_run_status(
    run: BaseRun,
    condition: V1StatusCondition,
    additional_fields: Optional[List[str]] = None,
    force: bool = False,
):
    previous_status = new_status(
        entity=run,
        condition=condition,
        additional_fields=additional_fields,
        force=force,
    )
    # Do not audit the new status since it's the same as the previous one
    if (
        condition.type in {V1Statuses.CREATED, V1Statuses.STOPPING}
        or previous_status == run.status
    ):
        return

    auditor.record(
        event_type=RUN_NEW_STATUS, instance=run, previous_status=previous_status
    )
    if run.status == V1Statuses.STOPPED:
        auditor.record(
            event_type=RUN_STOPPED, instance=run, previous_status=previous_status
        )
    elif run.status == V1Statuses.FAILED:
        auditor.record(
            event_type=RUN_FAILED, instance=run, previous_status=previous_status
        )
    elif run.status == V1Statuses.SUCCEEDED:
        auditor.record(
            event_type=RUN_SUCCEEDED, instance=run, previous_status=previous_status
        )
    elif run.status == V1Statuses.SKIPPED:
        auditor.record(
            event_type=RUN_SKIPPED, instance=run, previous_status=previous_status
        )
    elif run.status == V1Statuses.RESUMING:
        auditor.record(event_type=RUN_RESUMED, instance=run)

    # handle done status
    if LifeCycle.is_done(run.status):
        auditor.record(
            event_type=RUN_DONE, instance=run, previous_status=previous_status
        )


def new_run_stop_status(
    run: Models.Run, message: str, meta_info: Optional[Dict] = None
):
    # Update run status to show that its stopped
    message = f"Run is stopped; {message}" if message else "Run is stopped"
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.STOPPED,
        status="True",
        reason="StateManager",
        message=message,
        meta_info=meta_info,
    )
    new_run_status(run=run, condition=condition)


def new_run_stopping_status(
    run: Models.Run, message: str, meta_info: Optional[Dict] = None
) -> bool:
    if LifeCycle.is_done(run.status, progressing=True):
        return False

    if LifeCycle.is_safe_stoppable(run.status):
        new_run_stop_status(run, message, meta_info=meta_info)
        return True
    message = f"Run is stopping; {message}" if message else "Run is stopping"
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.STOPPING,
        status="True",
        reason="StateManager",
        message=message,
        meta_info=meta_info,
    )
    new_run_status(run=run, condition=condition)
    return True


def new_run_skip_status(
    run: Models.Run, message: str, meta_info: Optional[Dict] = None
):
    # Update run status to show that its stopped
    message = f"Run is skipped; {message}" if message else "Run is skipped"
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.SKIPPED,
        status="True",
        reason="StateManager",
        message=message,
        meta_info=meta_info,
    )
    new_run_status(run=run, condition=condition)


def new_run_skipped_status(
    run: Models.Run, message: str, meta_info: Optional[Dict] = None
) -> bool:
    if LifeCycle.is_done(run.status, progressing=True):
        return False

    if not LifeCycle.is_safe_stoppable(run.status):
        return False
    new_run_skip_status(run, message, meta_info)
    return True



from rest_framework import status
from rest_framework.generics import get_object_or_404
from rest_framework.response import Response

from django.conf import settings

from haupt.common.endpoints.base import BaseEndpoint, DestroyEndpoint, PostEndpoint
from haupt.db.defs import Models
from haupt.db.managers.bookmarks import bookmark_obj


class BookmarkCreateView(BaseEndpoint, PostEndpoint):
    """Base Bookmark create view."""

    queryset = None
    content_type = None

    def create(self, request, *args, **kwargs):
        obj = self.get_object()
        user_filter = {
            "user": self.request.user if settings.HAS_ORG_MANAGEMENT else None
        }
        bookmark_obj(**user_filter, obj=obj, content_type=self.content_type)
        self.audit(request, *args, **kwargs)
        return Response(status=status.HTTP_201_CREATED, data={})


class BookmarkDeleteView(BaseEndpoint, DestroyEndpoint):
    """Base Bookmark delete view."""

    queryset = None
    content_type = None

    def destroy(self, request, *args, **kwargs):
        user_filter = {"user": self.request.user} if settings.HAS_ORG_MANAGEMENT else {}
        obj = self.get_object()
        bookmark = get_object_or_404(
            Models.Bookmark,
            **user_filter,
            content_type__model=self.content_type,
            object_id=obj.id
        )
        bookmark.enabled = False
        bookmark.save(update_fields=["enabled"])
        self.audit(request, *args, **kwargs)
        return Response(status=status.HTTP_204_NO_CONTENT)



from typing import Dict, List, Optional, Tuple

from haupt.common.settings.apps import set_apps
from haupt.common.settings.assets import set_assets
from haupt.common.settings.celery import set_celery
from haupt.common.settings.core import set_core
from haupt.common.settings.cors import set_cors
from haupt.common.settings.middlewares import set_middlewares
from haupt.common.settings.ui import set_ui
from haupt.schemas.platform_config import PlatformConfig


def set_api_service(
    context,
    config: PlatformConfig,
    api_apps: Tuple,
    routes: Dict,
    db_app: Optional[str] = None,
    processors: Optional[List[str]] = None,
    enable_crsf: bool = False,
):
    project_apps = (
        "haupt.common.apis.apps.CommonApisConfig",
        "haupt.common.commands.apps.CommandsConfig",
    ) + api_apps
    set_apps(
        context=context,
        config=config,
        third_party_apps=("rest_framework", "corsheaders"),
        project_apps=project_apps,
        db_app=db_app,
        use_db_apps=True,
        use_admin_apps=True,
        use_staticfiles_app=True,
    )
    set_core(context=context, config=config, use_db=True)
    set_cors(context=context, config=config)
    set_ui(context=context, config=config, processors=processors)
    set_middlewares(context=context, config=config, enable_crsf=enable_crsf)
    set_assets(context=context, config=config)
    if config.scheduler_enabled:
        set_celery(context=context, config=config, routes=routes)


def set_service(context, config: PlatformConfig):
    from haupt.background.celeryp.routes import get_routes

    api_apps = (
        "haupt.apis.apps.APIsConfig",
        "haupt.streams.apps.StreamsConfig",
    )
    set_api_service(
        context, config, api_apps=api_apps, routes=get_routes(), db_app=None
    )



import os

from haupt.schemas.sandbox_config import SandboxConfig
from polyaxon._config.manager import ConfigManager
from polyaxon._config.spec import ConfigSpec


class SandboxConfigManager(ConfigManager):
    """Manages sandbox configuration .sandbox file."""

    VISIBILITY = ConfigManager.Visibility.ALL
    CONFIG_FILE_NAME = ".sandbox"
    CONFIG = SandboxConfig
    PERSIST_FORMAT = "yaml"

    @classmethod
    def get_config_from_env(cls):
        glob_path = cls.get_global_config_path()
        config_paths = [
            ConfigSpec(glob_path, config_type=".yaml", check_if_exists=False),
            os.environ,
            {"dummy": "dummy"},
        ]

        config = cls._CONFIG_READER.read_configs(config_paths)
        config = cls.CONFIG.from_dict(config.data)
        return config

    @classmethod
    def get_config_or_default(cls) -> SandboxConfig:
        if not cls.is_initialized():
            return cls.get_config_from_env()

        return cls.get_config()

    @classmethod
    def get_db_filepath(cls):
        path = cls.get_config_filepath()
        return os.path.dirname(path)



from haupt.common import auditor
from haupt.common.events.registry import component_version

auditor.subscribe(component_version.ComponentVersionDeletedEvent)
auditor.subscribe(component_version.ComponentVersionCreatedActorEvent)
auditor.subscribe(component_version.ComponentVersionUpdatedActorEvent)
auditor.subscribe(component_version.ComponentVersionViewedActorEvent)
auditor.subscribe(component_version.ComponentVersionDeletedActorEvent)
auditor.subscribe(component_version.ComponentVersionTransferredActorEvent)
auditor.subscribe(component_version.ComponentVersionNewStageEvent)



from haupt.common.events import event_actions, event_subjects
from haupt.common.events.event import ActorEvent, ActorTopEntityEvent
from haupt.common.events.registry.attributes import (
    PROJECT_OWNER_ATTRIBUTES,
    PROJECT_RESOURCE_OWNER_ATTRIBUTES,
)

PROJECT_BOOKMARKED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.BOOKMARKED, event_subjects.ACTOR
)
PROJECT_UNBOOKMARKED_ACTOR = "{}.{}.{}".format(
    event_subjects.PROJECT, event_actions.UNBOOKMARKED, event_subjects.ACTOR
)
RUN_BOOKMARKED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.BOOKMARKED, event_subjects.ACTOR
)
RUN_UNBOOKMARKED_ACTOR = "{}.{}.{}".format(
    event_subjects.RUN, event_actions.UNBOOKMARKED, event_subjects.ACTOR
)

EVENTS = {
    PROJECT_BOOKMARKED_ACTOR,
    PROJECT_UNBOOKMARKED_ACTOR,
    RUN_BOOKMARKED_ACTOR,
    RUN_UNBOOKMARKED_ACTOR,
}


class ProjectBookmarkedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_BOOKMARKED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class ProjectUnBookmarkedActorEvent(ActorTopEntityEvent):
    event_type = PROJECT_UNBOOKMARKED_ACTOR
    attributes = PROJECT_OWNER_ATTRIBUTES


class RunBookmarkedActorEvent(ActorEvent):
    event_type = RUN_BOOKMARKED_ACTOR
    entity_uuid = "project.uuid"
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES


class RunUnBookmarkedActorEvent(ActorEvent):
    event_type = RUN_UNBOOKMARKED_ACTOR
    entity_uuid = "project.uuid"
    attributes = PROJECT_RESOURCE_OWNER_ATTRIBUTES



import os

from typing import Optional

from clipped.utils.bools import to_bool

from haupt.cli.runners.base import start_app
from polyaxon._env_vars.keys import (
    ENV_KEYS_PROXY_GATEWAY_CONCURRENCY,
    ENV_KEYS_PROXY_GATEWAY_PER_CORE,
    ENV_KEYS_PROXY_STREAMS_TARGET_PORT,
)
from polyaxon._services.values import PolyaxonServices


def start(
    host: Optional[str] = None,
    port: Optional[int] = None,
    log_level: Optional[str] = None,
    workers: Optional[int] = None,
    per_core: bool = False,
    uds: Optional[str] = None,
):
    port = port or os.environ.get(ENV_KEYS_PROXY_STREAMS_TARGET_PORT)
    workers = workers or os.environ.get(ENV_KEYS_PROXY_GATEWAY_CONCURRENCY)
    per_core = per_core or to_bool(
        os.environ.get(ENV_KEYS_PROXY_GATEWAY_PER_CORE), handle_none=True
    )
    start_app(
        app="haupt.polyconf.asgi.streams:application",
        app_name=PolyaxonServices.STREAMS,
        host=host,
        port=port,
        log_level=log_level,
        workers=workers,
        per_core=per_core,
        uds=uds,
    )



from typing import Optional

from clipped.utils.imports import import_string

import click


@click.command()
@click.option(
    "--app",
    default="haupt",
)
@click.option("--service", default="all")
def queues(app: Optional[str], service: Optional[str]):
    app = app or "haupt"
    queues_manager = import_string(f"{app}.background.celeryp.queues.CeleryQueues")
    print(getattr(queues_manager, service)())



import datetime

from collections import namedtuple
from datetime import timedelta
from typing import Any, Optional

from django.utils import timezone

from haupt.common.options.option_owners import OptionOwners


class CachedOptionSpec(namedtuple("CachedOptionSpec", "value datetime")):
    pass


class MemoryCacheManager:
    INVALIDED_OPTION = "INVALIDED_OPTION"

    def __init__(self):
        self._state = {}

    def clear_key(self, key: str, owners: Optional[OptionOwners] = None):
        if owners:
            key = f"{key}:{owners}"
        self._state.pop(key, None)

    def clear(self):
        self._state = {}

    @classmethod
    def is_valid_value(cls, value: Any):
        return value != cls.INVALIDED_OPTION

    @staticmethod
    def is_valid_cache(value_datetime: datetime.datetime) -> bool:
        return timezone.now() < value_datetime

    def get_from_cache(self, key: str, owners: Optional[OptionOwners] = None) -> Any:
        if owners:
            key = f"{key}:{owners}"
        cached_option = self._state.get(key)
        if cached_option and self.is_valid_cache(cached_option.datetime):
            return cached_option.value
        self.clear_key(key=key)
        return self.INVALIDED_OPTION

    def set_to_cache(
        self, key: str, value: Any, ttl: int, owners: Optional[OptionOwners] = None
    ):
        if ttl <= 0 or value is None:
            return
        if owners:
            key = f"{key}:{owners}"
        self._state[key] = CachedOptionSpec(
            value=value, datetime=timezone.now() + timedelta(seconds=ttl)
        )



import os

from typing import Dict

from haupt.schemas.platform_config import PlatformConfig


def set_logging(
    context,
    config: PlatformConfig,
) -> Dict:
    context["LOG_DIRECTORY"] = config.logs_root
    if not os.path.exists(config.logs_root):
        os.makedirs(config.logs_root)
    logging_spec = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "[%(asctime)s] %(levelname)s %(message)s [%(name)s:%(lineno)s]",
                "datefmt": "%d/%b/%Y %H:%M:%S",
            },
            "simple": {"format": "%(levelname)8s  %(message)s [%(name)s]"},
            "verbose": {
                "format": "%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s"
            },
        },
        "filters": {
            "require_debug_false": {"()": "django.utils.log.RequireDebugFalse"}
        },
        "handlers": {
            "logfile": {
                "level": config.log_level,
                "class": "logging.handlers.RotatingFileHandler",
                "filename": "{}/polyaxon_{}.log".format(config.logs_root, os.getpid()),
                "maxBytes": 1024 * 1024 * 8,  # 8 MByte
                "backupCount": 5,
                "formatter": "standard",
            },
            "console": {
                "level": config.log_level,
                "class": "logging.StreamHandler",
                "formatter": "verbose",
            },
        },
        "loggers": {
            "django.request": {
                "propagate": True,
                "handlers": ["console"],
                "level": config.log_level,
            },
            "root": {"handlers": ["console"], "level": config.log_level},
        },
    }
    if config.is_debug_mode:
        logging_spec["loggers"]["django.db.backends"] = {
            "propagate": True,
            "handlers": ["console"],
            "level": config.log_level,
        }
    context["LOGGING"] = logging_spec



from urllib.parse import urlparse


def get_service_url(host: str, port: int):
    if port == 80:
        return "http://{}".format(host)
    if port == 443:
        return "https://{}".format(host)
    return "http://{}:{}".format(host, port)


def has_https(url: str):
    return "https" in url


def get_ssl_server_name(url: str):
    if has_https(url):
        return "proxy_ssl_server_name on;"
    return ""


def get_header_host(url: str):
    if has_https(url):
        return "proxy_set_header Host {};".format(urlparse(url).netloc.split(":")[0])
    return "proxy_set_header Host $http_host;"


def get_service_proxy(port: int = 8443):
    return "https://127.0.0.1:{}".format(port)



from haupt.managers.platform import PlatformManager
from polyaxon.settings import HOME_CONFIG

PlatformManager.set_config_path(HOME_CONFIG.path)
PLATFORM_CONFIG = PlatformManager.get_config_from_env(file_path=__file__)



from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

from polyaxon.schemas import V1Stages


class StageModel(models.Model):
    stage = models.CharField(
        max_length=16,
        blank=True,
        null=True,
        db_index=True,
        default=V1Stages.TESTING,
        choices=V1Stages.to_choices(),
    )
    stage_conditions = models.JSONField(
        encoder=DjangoJSONEncoder, blank=True, null=True, default=dict
    )

    class Meta:
        abstract = True



from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models

from polyaxon.schemas import V1RunEdgeKind, V1Statuses

if settings.DB_ENGINE_NAME == "sqlite":
    ArrayField = None
else:
    from django.contrib.postgres.fields import ArrayField


class BaseRunEdge(models.Model):
    downstream = models.ForeignKey(
        "db.Run", on_delete=models.CASCADE, related_name="upstream_edges"
    )
    upstream = models.ForeignKey(
        "db.Run", on_delete=models.CASCADE, related_name="downstream_edges"
    )
    values = models.JSONField(blank=True, null=True)
    kind = models.CharField(
        max_length=6,
        db_index=True,
        choices=V1RunEdgeKind.to_choices(),
        blank=True,
        null=True,
    )
    statuses = (
        models.JSONField(
            encoder=DjangoJSONEncoder,
            blank=True,
            null=True,
        )
        if settings.DB_ENGINE_NAME == "sqlite"
        else ArrayField(
            models.CharField(
                max_length=16,
                choices=V1Statuses.to_choices(),
            ),
            blank=True,
            null=True,
        )
    )

    class Meta:
        app_label = "db"
        db_table = "db_runedge"
        abstract = True



class CachedMixin:
    """
    A mixin to help clear cached properties.
    """

    CACHED_PROPERTIES = ()

    def clear_cached_properties(self, properties=None):
        properties = properties or self.CACHED_PROPERTIES
        for key in properties:
            if key in self.__dict__:
                del self.__dict__[key]



import logging

from haupt import settings
from haupt.managers.sandbox import SandboxConfigManager
from haupt.schemas.platform_config import PlatformConfig


def set_core(context, config: PlatformConfig, use_db: bool = True):
    context["DEBUG"] = config.is_debug_mode
    context["POLYAXON_SERVICE"] = config.service
    context["POLYAXON_ENVIRONMENT"] = config.env
    context["CHART_VERSION"] = config.chart_version
    context["SCHEDULER_ENABLED"] = config.scheduler_enabled
    context["K8S_NAMESPACE"] = config.namespace

    context["FILE_UPLOAD_PERMISSIONS"] = 0o644
    context[
        "MIN_ARTIFACTS_DELETION_TIMEDELTA"
    ] = config.min_artifacts_deletion_timedelta
    context["TTL_HEARTBEAT"] = config.ttl_heartbeat

    context["WSGI_APPLICATION"] = "{}.wsgi.application".format(config.config_module)
    context["TIME_ZONE"] = config.timezone
    context["LANGUAGE_CODE"] = "en"
    context["LANGUAGES"] = (("en", "English"),)

    context["USE_I18N"] = True
    context["USE_TZ"] = True

    context["INTERNAL_IPS"] = ("127.0.0.1",)
    context["APPEND_SLASH"] = True

    context["ROOT_URLCONF"] = ""

    if use_db:
        db_engine = (
            "django.db.backends.sqlite3"
            if config.is_sqlite_db_engine
            else "django.db.backends.postgresql"
        )
        context["AUTH_USER_MODEL"] = "db.User"
        context["DB_ENGINE_NAME"] = config.db_engine_name
        context["DEFAULT_DB_ENGINE"] = db_engine
        db_name = config.db_name
        if not db_name:
            if config.is_sqlite_db_engine:
                db_path = (
                    SandboxConfigManager.get_db_filepath()
                    if settings.SANDBOX_CONFIG
                    else "/tmp"
                )
                db_name = f"{db_path}/plx.db"
            else:
                db_name = "polyaxon"
        db_definition = {
            "ENGINE": db_engine,
            "NAME": db_name,
            "USER": config.db_user,
            "PASSWORD": config.db_password,
            "HOST": config.db_host,
            "PORT": config.db_port,
            "ATOMIC_REQUESTS": True,
            "CONN_MAX_AGE": config.db_conn_max_age,
        }
        if config.db_options:
            db_definition["OPTIONS"] = config.db_options
        context["DATABASES"] = {"default": db_definition}



from haupt.db.abstracts.artifacts import BaseArtifact, BaseArtifactLineage


class Artifact(BaseArtifact):
    class Meta(BaseArtifact.Meta):
        unique_together = (("name", "state"),)


class ArtifactLineage(BaseArtifactLineage):
    pass



from django.core.cache import cache


class SingletonMixin:
    """A base model to represents a singleton."""

    def set_cache(self):
        cache.set(self.__class__.__name__, self)

    def save(self, *args, **kwargs):  # pylint:disable=arguments-differ
        self.pk = 1
        super().save(*args, **kwargs)
        self.set_cache()

    def delete(self, *args, **kwargs):  # pylint:disable=arguments-differ
        pass

    @classmethod
    def load(cls):
        raise NotImplementedError



from clipped.utils.tz import get_datetime_from_now

from django.db.models import Count, Q

from haupt.background.celeryp.tasks import SchedulerCeleryTasks
from haupt.common import workers
from haupt.db.defs import Models
from haupt.db.managers.projects import update_project_based_on_last_updated_entities
from haupt.db.managers.runs import get_stopping_pipelines_with_no_runs
from haupt.db.managers.statuses import bulk_new_run_status
from polyaxon.schemas import LifeCycle, V1RunKind, V1StatusCondition, V1Statuses


class CronsHeartbeatManager:
    @staticmethod
    def heartbeat_out_of_sync_schedules():
        last_date = get_datetime_from_now(days=0, minutes=2)
        distressed_schedules = (
            Models.Run.objects.filter(
                status=V1Statuses.RUNNING,
                kind=V1RunKind.SCHEDULE,
            )
            .exclude(
                pipeline_runs__finished_at__gte=last_date,
            )
            .annotate(
                unfinished=Count(
                    "pipeline_runs",
                    filter=~Q(pipeline_runs__status__in=LifeCycle.DONE_VALUES),
                    distinct=True,
                ),
            )
            .filter(
                unfinished=0,
            )
            .values_list("id", flat=True)
        )
        for _id in distressed_schedules:
            workers.send(
                SchedulerCeleryTasks.RUNS_WAKEUP_SCHEDULE,
                kwargs={"run_id": _id},
            )

    @staticmethod
    def heartbeat_stopping_runs(minutes: int = 30):
        last_date = get_datetime_from_now(days=0, minutes=minutes)
        condition = V1StatusCondition.get_condition(
            type=V1Statuses.STOPPED,
            status="True",
            reason="HeartbeatStopping",
            message="Run is stopped by heartbeat process.",
        )

        # Finish all runs
        runs = Models.Run.objects.filter(
            status=V1Statuses.STOPPING,
            updated_at__lte=last_date,
            kind__in=[
                V1RunKind.JOB,
                V1RunKind.SERVICE,
                V1RunKind.TUNER,
                V1RunKind.NOTIFIER,
            ],
        )
        bulk_new_run_status(runs, condition)

        # Finish all stopping pipelines
        runs = get_stopping_pipelines_with_no_runs(
            Models.Run.objects.filter(updated_at__lte=last_date)
        )
        bulk_new_run_status(runs, condition)

    @staticmethod
    def heartbeat_project_last_updated(
        created_threshold: float = 2, updated_threshold: float = 15
    ):
        time_threshold = get_datetime_from_now(days=0, minutes=created_threshold)
        project_threshold = get_datetime_from_now(days=0, seconds=updated_threshold)
        update_project_based_on_last_updated_entities(
            last_created_at_threshold=time_threshold,
            last_updated_at_threshold=project_threshold,
        )



from clipped.utils.bools import to_bool
from rest_framework import serializers

from django.conf import settings

from haupt.common.apis.filters import OrderingFilter, QueryFilter
from haupt.common.authentication.base import is_user
from haupt.db.defs import Models
from haupt.db.query_managers.bookmarks import filter_bookmarks, get_bookmarks_filter


class BookmarkedSerializerMixin(serializers.Serializer):
    bookmarked_model = None

    bookmarked = serializers.SerializerMethodField()

    def get_bookmarked(self, obj):
        bookmarks = self.context.get("bookmarks", None)

        if bookmarks is not None:
            return obj.id in bookmarks
        else:
            # Get the requesting user if set in the context
            request = self.context.get("request", None)
            if not request:
                return False
            user = request.user
            if settings.HAS_ORG_MANAGEMENT and not is_user(user):
                return False
            user_filters = {"user": user} if settings.HAS_ORG_MANAGEMENT else {}
            if request:
                return Models.Bookmark.objects.filter(
                    **user_filters,
                    content_type__model=self.bookmarked_model,
                    object_id=obj.id,
                    enabled=True,
                ).exists()
        return False


class BookmarksQueryFilter(QueryFilter):
    def filter_queryset(self, request, queryset, view):
        queryset = filter_bookmarks(queryset, request, view)
        return super().filter_queryset(request, queryset, view)


class BookmarkedListMixinView:
    bookmarked_model = None
    filter_backends = (BookmarksQueryFilter, OrderingFilter)

    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class()
        if not issubclass(serializer_class, BookmarkedSerializerMixin):
            return super().get_serializer(*args, **kwargs)

        queryset = args[0]

        if not queryset or (
            settings.HAS_ORG_MANAGEMENT and not is_user(self.request.user)
        ):
            return super().get_serializer(*args, **kwargs)

        object_ids = [o.id for o in queryset]

        bookmarked_model, bookmarks_filter, has_bookmarks_filter = get_bookmarks_filter(
            self.request, self
        )
        if has_bookmarks_filter:
            # The queryset will be already filtered/excluded all bookmarks
            bookmarks = object_ids if to_bool(bookmarks_filter) else []
        else:
            # Batch-get all the bookmarks for the objects in the queryset
            # and pass them on to the serializer
            user_filters = (
                {"user": self.request.user} if settings.HAS_ORG_MANAGEMENT else {}
            )
            bookmarks = Models.Bookmark.objects.filter(
                **user_filters,
                content_type__model=serializer_class.bookmarked_model,
                object_id__in=object_ids,
                enabled=True,
            ).values_list("object_id", flat=True)

        context = self.get_serializer_context()
        context["bookmarks"] = list(bookmarks)
        kwargs["context"] = context
        return serializer_class(*args, **kwargs)



from django.db import migrations, models


def migrate_runtime(apps, schema_editor):
    Run = apps.get_model("db", "Run")

    runs = []
    for r in Run.objects.all():
        r.runtime = r.meta_info.pop("meta_kind", None)
        runs.append(r)

    Run.objects.bulk_update(runs, ["meta_info", "runtime"])


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0005_auto_20201005_0913"),
    ]

    operations = [
        migrations.AddField(
            model_name="run",
            name="runtime",
            field=models.CharField(blank=True, db_index=True, max_length=12, null=True),
        ),
        migrations.RunPython(migrate_runtime),
    ]



import os

from clipped.utils.paths import check_or_create_path, delete_path, untar_file

from asgiref.sync import sync_to_async
from polyaxon import settings
from polyaxon._fs.async_manager import upload_dir, upload_file
from polyaxon._fs.types import FSSystem
from polyaxon._fs.utils import get_store_path
from polyaxon.schemas import V1ProjectFeature


async def handle_posted_data(
    fs: FSSystem,
    store_path: str,
    content_file: any,
    root_path: str,
    path: str,
    upload: bool,
    is_file: bool,
    overwrite: bool = True,
    untar: bool = True,
) -> str:
    filename = content_file.name
    tmp_path = "{}/{}".format(root_path, os.path.basename(filename)).rstrip("/")
    if path:
        root_path = "{}/{}".format(root_path, path).rstrip("/")
        if is_file:
            root_path = "{}/{}".format(root_path, os.path.basename(filename))
    else:
        if untar:
            root_path = "{}/{}".format(root_path, path or "/").rstrip("/")
        else:
            root_path = tmp_path
    if not untar:
        tmp_path = root_path

    full_tmppath = settings.AGENT_CONFIG.get_local_path(
        subpath=tmp_path, entity=V1ProjectFeature.RUNTIME
    )
    full_filepath = settings.AGENT_CONFIG.get_local_path(
        subpath=root_path, entity=V1ProjectFeature.RUNTIME
    )
    store_full_path = get_store_path(
        store_path=store_path, subpath=root_path, entity=V1ProjectFeature.RUNTIME
    )

    if overwrite and os.path.exists(full_filepath):
        delete_path(full_filepath)
    if not overwrite and os.path.exists(full_filepath):
        return full_filepath
    # Always clean tmp path
    if overwrite and os.path.exists(full_tmppath):
        delete_path(full_tmppath)

    check_or_create_path(full_tmppath, is_dir=False)
    check_or_create_path(full_filepath, is_dir=not is_file)

    # Creating the new file
    with open(full_tmppath, "wb") as destination:
        for chunk in content_file.chunks():
            destination.write(chunk)
    if untar:
        await sync_to_async(untar_file)(
            full_tmppath, extract_path=full_filepath, use_filepath=False
        )
    if upload and store_full_path != full_filepath:
        if is_file:
            await upload_file(fs=fs, store_path=store_path, subpath=root_path)
        else:
            await upload_dir(fs=fs, store_path=store_path, subpath=root_path)
    return root_path



import os

from haupt.schemas.platform_config import PlatformConfig
from polyaxon.api import STATIC_V1


def set_assets(context, config: PlatformConfig):
    context["MEDIA_ROOT"] = config.media_root
    context["MEDIA_URL"] = config.media_url

    context["STATIC_ROOT"] = config.static_root or str(config.root_dir / "static")
    context["STATIC_URL"] = config.static_url or "/static/"

    # Additional locations of static files
    public_dir = str(config.root_dir / "public")
    if os.path.exists(public_dir):
        context["STATICFILES_DIRS"] = (public_dir,)

    context["STATICFILES_FINDERS"] = (
        "django.contrib.staticfiles.finders.FileSystemFinder",
        "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    )

    context["LOCALE_PATHS"] = (
        str(config.root_dir / "locale"),
        str(config.root_dir / "client" / "js" / "libs" / "locale"),
    )

    context["STATICI18N_ROOT"] = STATIC_V1
    context["STATICI18N_OUTPUT_DIR"] = "jsi18n"

    context["ARTIFACTS_ROOT"] = config.artifacts_root or "/tmp/plx/artifacts_uploads"
    context["ARCHIVES_ROOT"] = config.archives_root or "/tmp/plx/archives"



from django.conf import settings
from django.http import Http404, HttpRequest
from django.shortcuts import get_object_or_404

from haupt.apis.endpoints.project import ProjectResourceEndpoint
from haupt.common.apis.regex import (
    ARTIFACT_NAME_KEY,
    NAME_KEY,
    OWNER_NAME_KEY,
    PROJECT_NAME_KEY,
    RUN_UUID_KEY,
    UUID_KEY,
)
from haupt.common.permissions import PERMISSIONS_MAPPING
from haupt.db.defs import Models


class RunEndpoint(ProjectResourceEndpoint):
    queryset = (
        Models.Run.all.select_related("project", "project__owner")
        if settings.HAS_ORG_MANAGEMENT
        else Models.Run.all.select_related("project")
    )
    lookup_field = UUID_KEY
    lookup_url_kwarg = RUN_UUID_KEY
    permission_classes = PERMISSIONS_MAPPING.get(["RUN_PERMISSION"])
    throttle_scope = "run"
    CONTEXT_KEYS = (OWNER_NAME_KEY, PROJECT_NAME_KEY, RUN_UUID_KEY)
    CONTEXT_OBJECTS = ("run",)

    def enrich_queryset(self, queryset):
        filters = {"project__name": self.project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["project__owner__name"] = self.owner_name
        return queryset.filter(**filters)

    def set_owner(self):
        self.project = self.run.project
        self._owner_id = self.project.owner_id

    def initialize_object_context(self, request: HttpRequest, *args, **kwargs):
        #  pylint:disable=attribute-defined-outside-init
        self.run = self.get_object()


class RunResourceListEndpoint(RunEndpoint):
    permission_classes = PERMISSIONS_MAPPING.get(["RUN_PERMISSION"])
    AUDITOR_EVENT_TYPES = None

    def get_object(self):
        if self._object:
            return self._object

        filters = {"uuid": self.run_uuid, "project__name": self.project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["project__owner__name"] = self.owner_name
        self._object = get_object_or_404(
            Models.Run.restorable.select_related("project"),
            **filters,
        )

        # May raise a permission denied
        self.check_object_permissions(self.request, self._object)

        return self._object

    def enrich_queryset(self, queryset):
        return queryset.filter(run=self.run)


class RunArtifactEndpoint(RunEndpoint):
    lookup_field = NAME_KEY
    lookup_url_kwarg = ARTIFACT_NAME_KEY
    permission_classes = PERMISSIONS_MAPPING.get(["RUN_RESOURCE_PERMISSION"])
    AUDITOR_EVENT_TYPES = None
    CONTEXT_KEYS = (OWNER_NAME_KEY, PROJECT_NAME_KEY, RUN_UUID_KEY, ARTIFACT_NAME_KEY)
    CONTEXT_OBJECTS = ("run_artifact",)

    def enrich_queryset(self, queryset):
        filters = {"run__uuid": self.run_uuid, "run__project__name": self.project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["run__project__owner__name"] = self.owner_name
        return queryset.filter(**filters)

    def set_owner(self):
        self.project = self.run_artifact.run.project
        self._owner_id = self.project.owner_id

    def initialize_object_context(self, request: HttpRequest, *args, **kwargs):
        #  pylint:disable=attribute-defined-outside-init
        self.run_artifact = self.get_object()

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        try:
            return queryset.get(artifact__name=self.artifact_name)
        except Models.ArtifactLineage.DoesNotExist:
            raise Http404(
                "No %s matches the given query." % self.queryset.model._meta.object_name
            )



from typing import Optional, Tuple

from django.conf import settings

from haupt.common.options.exceptions import OptionException
from haupt.common.options.option import (
    NAMESPACE_DB_OPTION_MARKER,
    Option,
    OptionScope,
    OptionStores,
)
from haupt.common.options.option_namespaces import FEATURES


class Feature(Option):
    scope = OptionScope.USER
    is_secret = False
    is_optional = True
    is_list = False
    store = OptionStores(settings.STORE_OPTION)
    typing = "bool"
    default = True
    options = [True, False]
    immutable = False  # If immutable, the feature cannot be update by the user
    description = None

    @classmethod
    def get_marker(cls) -> str:
        return NAMESPACE_DB_OPTION_MARKER

    @classmethod
    def parse_key(cls) -> Tuple[Optional[str], str]:
        marker = cls.get_marker()
        parts = cls.key.split(marker)
        # First part is a Meta namespace `features`
        if len(parts) > 3 or len(parts) < 1:  # pylint:disable=len-as-condition
            raise OptionException(
                "Feature declared with multi-namespace key `{}`.".format(cls.key)
            )
        if parts[0] != FEATURES:
            raise OptionException(
                "Feature declared with wrong namespace key `{}`.".format(cls.key)
            )
        if len(parts) == 2:
            return None, parts[1]
        return parts[1], parts[2]

    @classmethod
    def get_namespace(cls) -> Optional[str]:
        return cls.parse_key()[0]

    @classmethod
    def get_key_subject(cls):
        return cls.parse_key()[1]



from typing import Optional


def write_to_conf_file(
    name: str, content: str, path: Optional[str] = None, ext: str = "conf"
):
    with open("{}/{}.{}".format(path or ".", name, ext), "w") as f:
        f.write(content)



from clipped.utils.bools import to_bool

from django.conf import settings

from haupt.db.defs import Models


def get_bookmarks_filter(request, view):
    bookmarked_model = getattr(view, "bookmarked_model")
    bookmarks = request.query_params.get("bookmarks")
    return (
        bookmarked_model,
        bookmarks,
        bookmarks is not None and bookmarked_model is not None,
    )


def filter_bookmarks(queryset, request, view):
    bookmarked_model, bookmarks, has_bookmarks_filter = get_bookmarks_filter(
        request, view
    )

    if (
        not has_bookmarks_filter
        or not request.user
        or (settings.HAS_ORG_MANAGEMENT and request.user.is_anonymous)
    ):
        return queryset

    user_filters = {"user": request.user} if settings.HAS_ORG_MANAGEMENT else {}
    bookmark_ids = Models.Bookmark.objects.filter(
        **user_filters,
        content_type__model=bookmarked_model,
        enabled=True,
    ).values_list("object_id", flat=True)

    if to_bool(bookmarks):
        return queryset.filter(id__in=bookmark_ids)
    else:
        return queryset.exclude(id__in=bookmark_ids)



from typing import Any, Mapping, Optional

from haupt.common.events.event import Event
from haupt.common.service_interface import Service


class EventService(Service):
    __all__ = ("record",)

    event_manager = None

    def can_handle(self, event_type: str) -> bool:
        return isinstance(event_type, str) and self.event_manager.knows(event_type)

    def get_event(
        self,
        event_type: str,
        event_data: Mapping = None,
        instance: Any = None,
        **kwargs
    ) -> Event:
        if instance or not event_data:
            return self.event_manager.get(event_type).from_instance(instance, **kwargs)
        return self.event_manager.get(event_type).from_event_data(
            event_data=event_data, **kwargs
        )

    def record(
        self,
        event_type: str,
        event_data: Mapping = None,
        instance: Any = None,
        **kwargs
    ) -> Optional[Event]:
        """Validate and record an event.

        >>> record('event.action', object_instance)
        """
        if not self.is_setup:
            return
        if not self.can_handle(event_type=event_type):
            return

        event = self.get_event(
            event_type=event_type, event_data=event_data, instance=instance, **kwargs
        )
        self.record_event(event)
        return event

    def record_event(self, event: Event):
        """Record an event.

        >>> record_event(Event())
        """



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models

from haupt.db.abstracts.diff import DiffModel


class BaseBookmark(DiffModel):
    enabled = models.BooleanField(default=True)
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    class Meta:
        abstract = True
        app_label = "db"
        db_table = "db_bookmark"
        verbose_name = "bookmark"
        verbose_name_plural = "bookmarks"

    def __str__(self) -> str:
        return "<{}-{}>".format(self.content_type, self.created_at)



from typing import Type

from clipped.decorators.cached_property import cached_property

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import models


class _Models:
    @staticmethod
    def get_db_model_name(name: str) -> str:
        db = settings.AUTH_USER_MODEL.split(".")[0]
        return "{}.{}".format(db, name)

    @classmethod
    def get_db_model(cls, name: str) -> Type[models.Model]:
        model_name = cls.get_db_model_name(name)
        try:
            return django_apps.get_model(model_name, require_ready=False)
        except ValueError:
            raise ImproperlyConfigured(
                "{} must be of the form 'app_label.model_name'".format(model_name)
            )
        except LookupError:
            raise ImproperlyConfigured(
                "{} refers to model '{}' that has not been installed".format(
                    name, model_name
                )
            )

    @cached_property
    def User(self) -> Type[models.Model]:
        return get_user_model()

    @cached_property
    def Project(self) -> Type[models.Model]:
        return self.get_db_model("Project")

    @cached_property
    def Run(self) -> Type[models.Model]:
        return self.get_db_model("Run")

    @cached_property
    def Artifact(self) -> Type[models.Model]:
        return self.get_db_model("Artifact")

    @cached_property
    def ArtifactLineage(self) -> Type[models.Model]:
        return self.get_db_model("ArtifactLineage")

    @cached_property
    def RunEdge(self) -> Type[models.Model]:
        return self.get_db_model("RunEdge")

    @cached_property
    def ProjectVersion(self) -> Type[models.Model]:
        return self.get_db_model("ProjectVersion")

    @cached_property
    def ProjectStats(self) -> Type[models.Model]:
        return self.get_db_model("ProjectStats")

    @cached_property
    def Bookmark(self) -> Type[models.Model]:
        return self.get_db_model("Bookmark")


Models = _Models()



from django.db import models

from polyaxon.schemas import LiveState


class LiveManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(live_state=LiveState.LIVE)


class ArchivedManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(live_state=LiveState.ARCHIVED)


class RestorableManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.filter(live_state__in={LiveState.LIVE, LiveState.ARCHIVED})



from typing import List, Optional

from haupt import settings
from haupt.proxies.schemas.buffering import get_buffering_config
from haupt.proxies.schemas.charset import get_charset_config
from haupt.proxies.schemas.error_page import get_error_page_config
from haupt.proxies.schemas.favicon import get_favicon_config
from haupt.proxies.schemas.gzip import get_gzip_config
from haupt.proxies.schemas.listen import get_listen_config
from haupt.proxies.schemas.logging import get_logging_config
from haupt.proxies.schemas.robots import get_robots_config
from haupt.proxies.schemas.services import (
    get_auth_request_config,
    get_internal_location_config,
    get_k8s_location_config,
    get_services_definitions,
    get_streams_location_config,
)
from haupt.proxies.schemas.ssl import get_ssl_config
from haupt.proxies.schemas.timeout import get_timeout_config


def get_scaffold_config(
    is_proxy: bool,
    port: int = 8000,
    use_ssl_config: bool = False,
    use_assets_config: bool = False,
    use_services_configs: bool = False,
    resolver: Optional[str] = None,
    auth: Optional[str] = None,
    api_configs: Optional[List[str]] = None,
    api_location_configs: Optional[List[str]] = None,
    is_local_streams_service: bool = False,
) -> List[str]:
    config = [get_listen_config(is_proxy=is_proxy, port=port)]
    if use_ssl_config and settings.PROXIES_CONFIG.ssl_enabled:
        config.append(get_ssl_config())
    config += [
        get_logging_config(),
        get_gzip_config(),
        get_charset_config(),
        get_buffering_config(),
        get_timeout_config(),
    ]
    if api_configs:
        config += api_configs
    config.append(get_error_page_config())
    if use_assets_config:
        config += [
            get_robots_config(),
            get_favicon_config(),
        ]
    if is_local_streams_service:
        config.append(
            get_auth_request_config(),
        )
    if use_services_configs:
        config += [
            get_internal_location_config(
                resolver=resolver, is_local_service=is_local_streams_service
            ),
            get_streams_location_config(
                resolver=resolver, auth=auth, is_local_service=is_local_streams_service
            ),
            get_k8s_location_config(
                resolver=resolver, auth=auth, is_local_service=is_local_streams_service
            ),
            # get_plugins_location_config(resolver=resolver, auth=auth)
        ]
        config += get_services_definitions(resolver=resolver, auth=auth)
    if api_location_configs:
        config += api_location_configs
    return config



from rest_framework import fields, serializers


class ProjectMixin(serializers.Serializer):
    project = fields.SerializerMethodField()

    def get_project(self, obj):
        return obj.project.name



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0003_run_pipeline"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="project",
            name="deleted",
        ),
        migrations.RemoveField(
            model_name="run",
            name="deleted",
        ),
        migrations.AddField(
            model_name="project",
            name="live_state",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "live"),
                    (0, "archived"),
                    (-1, "deletion_progressing"),
                ],
                default=1,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="run",
            name="live_state",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "live"),
                    (0, "archived"),
                    (-1, "deletion_progressing"),
                ],
                default=1,
                null=True,
            ),
        ),
    ]



from typing import Optional

from django.conf import settings

from haupt.common.authentication.base import is_normal_user
from haupt.db.managers.stages import new_stage
from haupt.db.managers.versions import add_version_contributors
from polyaxon.schemas import V1StageCondition


def create_stage(view, serializer, event_type: Optional[str] = None):
    serializer.is_valid()
    validated_data = serializer.validated_data
    if not validated_data:
        return
    condition = None
    if validated_data.get("condition"):
        condition = V1StageCondition.get_condition(**validated_data.get("condition"))
    if not condition:
        return
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(view.request.user):
        status_meta_info = {
            "user": {
                "username": view.request.user.username,
                "email": view.request.user.email,
            },
        }
        condition.meta_info = status_meta_info
    new_stage(
        entity=view.version,
        owner_id=view._owner_id,
        user=view.request.user,
        condition=condition,
        event_type=event_type,
    )
    add_version_contributors(view.version, users=[view.request.user])



from haupt.schemas.platform_config import PlatformConfig


def set_encryption(context, config: PlatformConfig):
    context["ENCRYPTION_KEY"] = config.encryption_key
    context["ENCRYPTION_SECRET"] = config.encryption_secret
    context["ENCRYPTION_BACKEND"] = config.encryption_backend



from haupt.schemas.platform_config import PlatformConfig


def set_middlewares(context, config: PlatformConfig, enable_crsf: bool = False):
    context["MIDDLEWARE"] = (
        "django.middleware.security.SecurityMiddleware",
        "django.contrib.sessions.middleware.SessionMiddleware",
        "corsheaders.middleware.CorsMiddleware",
        "django.middleware.common.CommonMiddleware",
    )

    if enable_crsf:
        context["MIDDLEWARE"] += ("django.middleware.csrf.CsrfViewMiddleware",)

    context["MIDDLEWARE"] += (
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "django.contrib.messages.middleware.MessageMiddleware",
        "django.middleware.clickjacking.XFrameOptionsMiddleware",
    )

    if context["UI_IN_SANDBOX"]:
        context["MIDDLEWARE"] += ("whitenoise.middleware.WhiteNoiseMiddleware",)


def set_base_middlewares(context, config: PlatformConfig):
    context["MIDDLEWARE"] = ("django.middleware.common.CommonMiddleware",)
    if context["UI_IN_SANDBOX"]:
        context["MIDDLEWARE"] += ("whitenoise.middleware.WhiteNoiseMiddleware",)



from django.core.validators import validate_slug
from django.db import models

from haupt.common.validation.blacklist import validate_blacklist_name
from haupt.db.abstracts.catalogs import BaseLiveStateCatalog
from haupt.db.abstracts.contributors import ContributorsModel
from haupt.db.abstracts.projects import Actor, Owner
from haupt.db.abstracts.readme import ReadmeModel


class Project(BaseLiveStateCatalog, ReadmeModel, ContributorsModel):
    latest_stats = models.OneToOneField(
        "db.ProjectStats",
        related_name="+",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    name = models.CharField(
        max_length=128, validators=[validate_slug, validate_blacklist_name], unique=True
    )

    class Meta(BaseLiveStateCatalog.Meta):
        app_label = "db"
        db_table = "db_project"

    @property
    def owner(self):
        return Owner

    @property
    def owner_id(self):
        return Owner.id

    @property
    def actor(self):
        return Actor



# isort: skip_file

# Default configs

from haupt.polyconf.config_manager import PLATFORM_CONFIG

from haupt.common.settings.logging import set_logging
from haupt.common.settings.admin import set_admin
from haupt.common.settings.secrets import set_secrets

context = locals()
MAX_CONCURRENCY = PLATFORM_CONFIG.max_concurrency or 1
set_logging(context=context, config=PLATFORM_CONFIG)
set_admin(context=context, config=PLATFORM_CONFIG)
set_secrets(context=context, config=PLATFORM_CONFIG)
if PLATFORM_CONFIG.is_streams_service:
    from haupt.common.settings.services.streams import set_service
elif PLATFORM_CONFIG.is_api_service:
    from haupt.common.settings.services.api import set_service
else:
    from haupt.common.settings.services.background import set_service

set_service(context=context, config=PLATFORM_CONFIG)

from haupt.common.settings.defaults import *

if PLATFORM_CONFIG.is_api_service or PLATFORM_CONFIG.is_streams_service:
    from .rest import *



from clipped.decorators.signals import ignore_raw, ignore_updates_pre

from django.db.models.signals import pre_save
from django.dispatch import receiver
from django.utils.timezone import now

from haupt.db.defs import Models


@receiver(pre_save, sender=Models.ProjectStats)
@ignore_updates_pre
@ignore_raw
def round_to_hour(sender, instance, **kwargs):
    instance.created_at = now().replace(minute=0, second=0, microsecond=0)



from collections import namedtuple
from typing import Dict, List, Optional, Set, Tuple, Union

from clipped.utils.lists import to_list

from haupt.common.service_interface import Service
from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from haupt.db.managers.statuses import new_run_status
from polyaxon._constants.metadata import (
    META_COPY_ARTIFACTS,
    META_DESTINATION_IMAGE,
    META_HAS_DAGS,
    META_HAS_HOOKS,
    META_HAS_JOBS,
    META_HAS_MATRICES,
    META_HAS_SERVICES,
    META_ITERATION,
    META_RECOMPILE,
    META_UPLOAD_ARTIFACTS,
)
from polyaxon._polyaxonfile import OperationSpecification
from polyaxon._schemas.types import V1ArtifactsType
from polyaxon.schemas import (
    ManagedBy,
    V1CloningKind,
    V1CompiledOperation,
    V1MatrixKind,
    V1Operation,
    V1RunEdgeKind,
    V1RunKind,
    V1RunPending,
    V1ScheduleKind,
    V1StatusCondition,
    V1Statuses,
)


class OperationInitSpec(
    namedtuple(
        "OperationInitSpec",
        "compiled_operation instance related_instance",
    )
):
    def update(self, compiled_operation=None, instance=None, related_instance=None):
        compiled_operation = compiled_operation or self.compiled_operation
        instance = instance or self.instance
        related_instance = related_instance or self.related_instance
        return OperationInitSpec(compiled_operation, instance, related_instance)


class OperationsService(Service):
    SUPPORTS_OP_BUILD = True
    DEFAULT_KINDS = V1RunKind.to_set() | V1MatrixKind.to_set() | V1ScheduleKind.to_set()

    __all__ = (
        "init_run",
        "init_and_save_run",
        "resolve_build",
        "save_build_relation",
        "resume_run",
        "restart_run",
        "copy_run",
    )

    @staticmethod
    def set_spec(spec: V1Operation, **kwargs) -> Tuple[V1Operation, Dict]:
        kwargs["raw_content"] = spec.to_json()
        return spec, kwargs

    @staticmethod
    def get_kind(compiled_operation: V1CompiledOperation) -> Tuple[str, Optional[str]]:
        if compiled_operation.is_tf_job_run:
            return V1RunKind.JOB, V1RunKind.TFJOB
        elif compiled_operation.is_pytorch_job_run:
            return V1RunKind.JOB, V1RunKind.PYTORCHJOB
        elif compiled_operation.is_paddle_job_run:
            return V1RunKind.JOB, V1RunKind.PADDLEJOB
        elif compiled_operation.is_mx_job_run:
            return V1RunKind.JOB, V1RunKind.MXJOB
        elif compiled_operation.is_xgb_job_run:
            return V1RunKind.JOB, V1RunKind.XGBJOB
        elif compiled_operation.is_mpi_job_run:
            return V1RunKind.JOB, V1RunKind.MPIJOB
        elif compiled_operation.is_ray_job_run:
            return V1RunKind.JOB, V1RunKind.RAYJOB
        elif compiled_operation.is_dask_job_run:
            return V1RunKind.JOB, V1RunKind.DASKJOB
        elif compiled_operation.is_tuner_run:
            return V1RunKind.JOB, V1RunKind.TUNER
        elif compiled_operation.is_notifier_run:
            return V1RunKind.JOB, V1RunKind.NOTIFIER
        # Default case
        kind = compiled_operation.run.kind
        return kind, kind

    @classmethod
    def supports_kind(
        cls, kind: str, runtime: str, supported_kinds: Set[str], is_managed: bool
    ) -> bool:
        supported_kinds = supported_kinds or cls.DEFAULT_KINDS
        error_message = (
            "You cannot create this operation. This can happen if "
            "the current project has runtime restrictions, "
            "your account has reached the allowed quota, "
            "or your plan does not support operations of kind: {}"
        )
        if kind not in supported_kinds and is_managed:
            raise ValueError(error_message.format(kind))
        if runtime and runtime not in supported_kinds and is_managed:
            raise ValueError(error_message.format(runtime))
        return True

    @classmethod
    def _finalize_meta_info(cls, meta_info: Dict, **kwargs):
        iteration = kwargs.pop(META_ITERATION, None)
        if iteration is not None:
            meta_info[META_ITERATION] = iteration
        return meta_info

    @classmethod
    def get_meta_info(
        cls,
        compiled_operation: V1CompiledOperation,
        kind: str,
        runtime: str,
        meta_info: Optional[Dict] = None,
        **kwargs,
    ) -> Tuple[str, str, Dict]:
        meta_info = meta_info or {}
        if compiled_operation.hooks:
            meta_info[META_HAS_HOOKS] = True
        if compiled_operation.schedule:
            if compiled_operation.matrix:
                meta_info[META_HAS_MATRICES] = True
            elif kind == V1RunKind.JOB:
                meta_info[META_HAS_JOBS] = True
            elif kind == V1RunKind.SERVICE:
                meta_info[META_HAS_SERVICES] = True
            elif kind == V1RunKind.DAG:
                meta_info[META_HAS_DAGS] = True
            kind = V1RunKind.SCHEDULE
            runtime = compiled_operation.schedule.kind
        elif compiled_operation.matrix:
            if kind == V1RunKind.JOB:
                meta_info[META_HAS_JOBS] = True
            elif kind == V1RunKind.SERVICE:
                meta_info[META_HAS_SERVICES] = True
            elif kind == V1RunKind.DAG:
                meta_info[META_HAS_DAGS] = True
            kind = V1RunKind.MATRIX
            runtime = compiled_operation.matrix.kind

        meta_info = cls._finalize_meta_info(meta_info=meta_info, **kwargs)
        return kind, runtime, meta_info

    @staticmethod
    def sanitize_kwargs(**kwargs):
        kwargs.pop(META_ITERATION, None)
        kwargs.pop("supported_owners", None)
        return kwargs

    def is_valid(self, compiled_operation: V1CompiledOperation):
        compiled_operation.validate_build()
        if compiled_operation.build and not self.SUPPORTS_OP_BUILD:
            raise ValueError(
                "You cannot create this operation. "
                "The build section is not supported in your plan."
            )

    def resolve_build(
        self,
        project_id: int,
        user_id: int,
        compiled_operation: V1CompiledOperation,
        inputs: Dict,
        meta_info: Optional[Dict] = None,
        pending: Optional[str] = None,
        **kwargs,
    ):
        instance = self.init_run(
            project_id=project_id,
            user_id=user_id,
            op_spec=V1Operation.from_build(compiled_operation.build, contexts=inputs),
            pending=pending,
            meta_info=meta_info,
            supported_owners=kwargs.get("supported_owners"),
        ).instance
        instance.runtime = V1RunKind.BUILDER
        return instance

    def _clone_build(self, original_run, run):
        build = (
            original_run.upstream_runs.filter(
                downstream_edges__kind=V1RunEdgeKind.BUILD
            )
            .order_by("created_at")
            .only("id")
            .last()
        )
        Models.RunEdge(
            upstream_id=build.id,
            downstream_id=run.id,
            kind=V1RunEdgeKind.BUILD,
        ).save()

    @staticmethod
    def save_build_relation(run_init_spec: OperationInitSpec):
        run_init_spec.related_instance.save()
        Models.RunEdge(
            upstream_id=run_init_spec.related_instance.id,
            downstream_id=run_init_spec.instance.id,
            kind=V1RunEdgeKind.BUILD,
        ).save()

    def init_run(
        self,
        project_id: int,
        user_id: int,
        op_spec: V1Operation = None,
        compiled_operation: V1CompiledOperation = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[str] = None,
        override: Optional[Union[str, Dict]] = None,
        use_override_patch_strategy: bool = False,
        params: Optional[Dict] = None,
        readme: Optional[str] = None,
        original_id: Optional[int] = None,
        cloning_kind: Optional[str] = None,
        managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
        pending: Optional[str] = None,
        meta_info: Optional[Dict] = None,
        supported_kinds: Set[str] = None,
        **kwargs,
    ) -> OperationInitSpec:
        if op_spec:
            op_spec, kwargs = self.set_spec(op_spec, **kwargs)
        if op_spec:
            if not compiled_operation or override:
                compiled_operation = OperationSpecification.compile_operation(
                    op_spec,
                    override=override,
                    use_override_patch_strategy=use_override_patch_strategy,
                )
            params = op_spec.params

        params = params or {}
        inputs = {p: pv.value for p, pv in params.items() if pv.is_literal}
        params = {p: pv.to_dict() for p, pv in params.items()}
        kind = None
        meta_info = meta_info or {}
        build_instance = None
        runtime = None
        if compiled_operation:
            self.is_valid(compiled_operation)
            # If the is ab upload we need to check the build process immediately
            if pending == V1RunPending.UPLOAD and compiled_operation.build:
                upload_artifacts = meta_info.pop(META_UPLOAD_ARTIFACTS, None)
                build_instance = self.resolve_build(
                    project_id=project_id,
                    user_id=user_id,
                    compiled_operation=compiled_operation,
                    inputs=inputs,
                    meta_info={META_UPLOAD_ARTIFACTS: upload_artifacts}
                    if upload_artifacts
                    else None,
                    pending=V1RunPending.UPLOAD,
                    **kwargs,
                )
                # Change the pending logic to wait to build and remove build requirements
                compiled_operation.build = None
                pending = V1RunPending.BUILD
            # If the is a clone/resume we need to check the build process and remove it
            if cloning_kind and compiled_operation.build:
                compiled_operation.build = None

            if pending is None and compiled_operation.is_approved is False:
                pending = V1RunPending.APPROVAL
            name = name or compiled_operation.name
            description = description or compiled_operation.description
            tags = tags or compiled_operation.tags
            if tags:
                tags = to_list(tags, check_none=True, to_unique=True)
            kind, runtime = self.get_kind(compiled_operation)
            kind, runtime, meta_info = self.get_meta_info(
                compiled_operation, kind, runtime, meta_info, **kwargs
            )
            self.supports_kind(
                kind, runtime, supported_kinds, ManagedBy.is_managed(managed_by)
            )
            kwargs["content"] = compiled_operation.to_json()
        instance = Models.Run(
            project_id=project_id,
            user_id=user_id,
            name=name,
            description=description,
            tags=tags,
            readme=readme,
            params=params,
            inputs=inputs,
            kind=kind,
            runtime=runtime,
            meta_info=meta_info,
            original_id=original_id,
            cloning_kind=cloning_kind,
            managed_by=managed_by,
            pending=pending,
            status_conditions=[
                V1StatusCondition.get_condition(
                    type=V1Statuses.CREATED,
                    status="True",
                    reason=kwargs.pop("reason", "OperationServiceInit"),
                    message=kwargs.pop("message", "Run is created"),
                ).to_dict()
            ],
            **self.sanitize_kwargs(**kwargs),
        )
        return OperationInitSpec(compiled_operation, instance, build_instance)

    def init_and_save_run(
        self,
        project_id: int,
        user_id: int,
        op_spec: V1Operation = None,
        compiled_operation: V1CompiledOperation = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[str] = None,
        override: Optional[Union[str, Dict]] = None,
        params: Optional[Dict] = None,
        readme: Optional[str] = None,
        managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
        pending: Optional[str] = None,
        meta_info: Optional[Dict] = None,
        pipeline_id: Optional[int] = None,
        controller_id: Optional[int] = None,
        supported_kinds: Set[str] = None,
        supported_owners: Set[str] = None,
    ):
        run_init_spec = self.init_run(
            project_id=project_id,
            user_id=user_id,
            name=name,
            description=description,
            op_spec=op_spec,
            compiled_operation=compiled_operation,
            override=override,
            params=params,
            readme=readme,
            pipeline_id=pipeline_id,
            controller_id=controller_id,
            tags=tags,
            managed_by=managed_by,
            pending=pending,
            meta_info=meta_info,
            supported_kinds=supported_kinds,
            supported_owners=supported_owners,
        )
        run_init_spec.instance.save()
        if run_init_spec.related_instance:
            self.save_build_relation(run_init_spec)
        return run_init_spec.instance

    def resume_run(
        self,
        run: BaseRun,
        user_id: Optional[int] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        content: Optional[str] = None,
        readme: Optional[str] = None,
        tags: Optional[List[str]] = None,
        supported_kinds: Set[str] = None,
        message=None,
        **kwargs,
    ):
        meta_info = kwargs.pop("meta_info", {}) or {}
        status_meta_info = kwargs.pop("status_meta_info", None)
        recompile = meta_info.pop(META_RECOMPILE, False)
        if recompile:
            op_spec = V1Operation.read(content)
            content = None
        else:
            op_spec = V1Operation.read(run.raw_content)  # TODO: Use constructor
        instance = self.init_run(
            project_id=run.project_id,
            user_id=user_id or run.user_id,
            name=name or run.name,
            description=description or run.description,
            readme=readme or run.readme,
            op_spec=op_spec,
            tags=tags or run.tags,
            override=content,
            supported_kinds=supported_kinds,
            cloning_kind=V1CloningKind.RESTART,  # To clean the build if any
            use_override_patch_strategy=True,
            **kwargs,
        ).instance

        run.user_id = instance.user_id
        run.name = instance.name
        run.description = instance.description
        run.readme = instance.readme
        run.content = instance.content
        run.raw_content = instance.raw_content
        run.tags = instance.tags
        run.params = instance.params
        run.save()
        status_condition = V1StatusCondition.get_condition(
            type=V1Statuses.RESUMING,
            status=True,
            reason="ResumeManager",
            message=message,
        )
        if status_meta_info:
            status_condition.meta_info = status_meta_info
        new_run_status(
            run,
            condition=status_condition,
            force=True,
        )
        return run

    def _clone_run(
        self,
        run: BaseRun,
        cloning_kind: str,
        user_id: Optional[int] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        content: Optional[str] = None,
        readme: Optional[str] = None,
        tags: List[int] = None,
        supported_kinds: Set[str] = None,
        supported_owners: Set[str] = None,
        **kwargs,
    ) -> BaseRun:
        meta_info = kwargs.pop("meta_info", {}) or {}
        recompile = meta_info.pop(META_RECOMPILE, False)
        if recompile:
            op_spec = V1Operation.read(content)
            content = None
        else:
            op_spec = V1Operation.read(run.raw_content)  # TODO: Use constructor
        original_meta_info = run.meta_info or {}
        original_uuid = run.uuid.hex
        upload_artifacts = original_meta_info.get(META_UPLOAD_ARTIFACTS)
        build_destination = original_meta_info.get(META_DESTINATION_IMAGE)
        if build_destination:
            meta_info[META_DESTINATION_IMAGE] = build_destination
        if upload_artifacts:
            meta_info[META_UPLOAD_ARTIFACTS] = upload_artifacts
        if cloning_kind == V1CloningKind.COPY and META_COPY_ARTIFACTS not in meta_info:
            # Handle default copy mode
            meta_info[META_COPY_ARTIFACTS] = V1ArtifactsType(
                dirs=[original_uuid]
            ).to_dict()
        if META_COPY_ARTIFACTS not in meta_info and upload_artifacts:
            # Handle default copy mode
            meta_info[META_COPY_ARTIFACTS] = V1ArtifactsType(
                dirs=["{}/{}".format(original_uuid, upload_artifacts)]
            ).to_dict()

        instance = self.init_run(
            project_id=run.project_id,
            user_id=user_id or run.user_id,
            name=name or run.name,
            description=description or run.description,
            readme=readme or run.readme,
            op_spec=op_spec,
            original_id=run.id,
            cloning_kind=cloning_kind,
            tags=tags or run.tags,
            override=content,
            supported_kinds=supported_kinds,
            supported_owners=supported_owners,
            meta_info=meta_info,
            use_override_patch_strategy=True,
            **kwargs,
        ).instance
        instance.save()
        if build_destination:
            self._clone_build(original_run=run, run=instance)
        return instance

    def restart_run(
        self,
        run: BaseRun,
        user_id: Optional[int] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        content: Optional[str] = None,
        readme: Optional[str] = None,
        tags: List[int] = None,
        supported_kinds: Set[str] = None,
        supported_owners: Set[str] = None,
        **kwargs,
    ) -> BaseRun:
        return self._clone_run(
            run=run,
            cloning_kind=V1CloningKind.RESTART,
            user_id=user_id,
            name=name,
            description=description,
            content=content,
            readme=readme,
            tags=tags,
            supported_kinds=supported_kinds,
            supported_owners=supported_owners,
            **kwargs,
        )

    def copy_run(
        self,
        run: BaseRun,
        user_id: Optional[int] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        content: Optional[str] = None,
        readme: Optional[str] = None,
        tags: List[int] = None,
        supported_kinds: Set[str] = None,
        supported_owners: Set[str] = None,
        **kwargs,
    ) -> BaseRun:
        return self._clone_run(
            run=run,
            cloning_kind=V1CloningKind.COPY,
            user_id=user_id,
            name=name,
            description=description,
            content=content,
            readme=readme,
            tags=tags,
            supported_kinds=supported_kinds,
            supported_owners=supported_owners,
            **kwargs,
        )



import uuid

from typing import List, Optional

from django.conf import settings

from haupt.common.authentication.base import is_normal_user
from haupt.db.abstracts.projects import Owner
from haupt.db.defs import Models
from polyaxon.pkg import SCHEMA_VERSION
from polyaxon.schemas import V1Component


def add_version_contributors(
    version: Models.ProjectVersion,
    users: Optional[List[Models.User]] = None,
    user_ids: Optional[List[int]] = None,
):
    if not settings.HAS_ORG_MANAGEMENT:
        return
    if not version:
        return
    _users = [u.id for u in users if is_normal_user(u)] if users else user_ids
    if not _users:
        return

    version.contributors.add(*_users)


def get_component_version_state(
    component: V1Component,
) -> Optional[uuid.UUID]:
    """A string representation that is used to create hash version"""
    component.kind = "component"
    component.version = SCHEMA_VERSION
    return uuid.uuid5(Owner.uuid, component.to_json())



from datetime import timedelta
from typing import Dict, List, Optional, Tuple

from django.conf import settings as dj_settings
from django.db.models import Count, Q
from django.utils.timezone import now

from haupt.background.celeryp.tasks import CronsCeleryTasks, SchedulerCeleryTasks
from haupt.common import workers
from haupt.db.defs import Models
from haupt.db.managers.live_state import confirm_delete_runs
from haupt.db.managers.queues import get_num_to_start
from haupt.db.managers.statuses import bulk_new_run_status
from haupt.db.queries.runs import STATUS_UPDATE_COLUMNS_ONLY
from polyaxon import _operations, settings
from polyaxon._auxiliaries import V1DefaultScheduling
from polyaxon._schemas.agent import AgentConfig
from polyaxon._utils.fqn_utils import get_run_instance
from polyaxon.schemas import (
    LifeCycle,
    LiveState,
    ManagedBy,
    V1RunKind,
    V1StatusCondition,
    V1Statuses,
)

MAX_DELETE_ITEMS = 200


def check_schedules(
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    agent_filters: Optional[Dict] = None,
) -> bool:
    agent_filters = agent_filters or {}
    end = now() + timedelta(seconds=3)
    schedule_runs = (
        Models.Run.objects.filter(
            **agent_filters,
            status=V1Statuses.CREATED,
            schedule_at__lte=end,
            pending__isnull=True,
            managed_by=managed_by,
        )
        .order_by("schedule_at")
        .only(*STATUS_UPDATE_COLUMNS_ONLY)
    )
    if schedule_runs:
        condition = V1StatusCondition.get_condition(
            type=V1Statuses.ON_SCHEDULE,
            status="True",
            reason="AgentSchedules",
            message="Run is on schedule",
        )
        bulk_new_run_status(runs=schedule_runs, condition=condition)
        schedule_run_ids = [s.id for s in schedule_runs]
        for run_id in schedule_run_ids:
            workers.send(
                SchedulerCeleryTasks.RUNS_PREPARE,
                kwargs={"run_id": run_id},
            )
    return Models.Run.objects.filter(
        **agent_filters,
        status=V1Statuses.CREATED,
        schedule_at__range=(end, end + timedelta(seconds=8)),
        pending__isnull=True,
        managed_by=managed_by,
    ).exists()


def get_stopping_runs(
    owner_name: str,
    max_budget: int,
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    agent_filters: Optional[Dict] = None,
    is_new_agent: Optional[bool] = False,
) -> Tuple[List[Tuple[str, str, str]], bool]:
    agent_filters = agent_filters or {}
    filters = Q(status=V1Statuses.STOPPING)
    if dj_settings.HAS_ORG_MANAGEMENT:
        values_list = ["project__name", "uuid", "kind", "namespace"]
    else:
        values_list = ["project__name", "uuid", "kind"]
    stopping_runs = (
        Models.Run.restorable.filter(
            **agent_filters,
            kind__in=[
                V1RunKind.JOB,
                V1RunKind.SERVICE,
                V1RunKind.TUNER,
                V1RunKind.NOTIFIER,
            ],
            pending__isnull=True,
            managed_by=managed_by,
        )
        .filter(filters)
        .prefetch_related("project")
        .values_list(*values_list)[:max_budget]
    )
    full = len(stopping_runs) >= max_budget
    if is_new_agent:
        data = [
            (
                get_run_instance(owner_name, run[0], run[1].hex),
                run[2],
                run[3] if dj_settings.HAS_ORG_MANAGEMENT else None,
            )
            for run in stopping_runs
        ]
    else:
        data = [
            (get_run_instance(owner_name, run[0], run[1].hex), run[2])
            for run in stopping_runs
        ]
    return data, full


def get_deleting_runs(
    owner_name: str,
    agent_id: str,
    agent_config: AgentConfig,
    max_budget: int,
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    agent_filters: Optional[Dict] = None,
    is_new_agent: Optional[bool] = None,
) -> Tuple[List[str], List[Tuple[str, str, str, str, str]], bool]:
    agent_filters = agent_filters or {}
    values = []

    # Clean and delete
    deleting_runs = (
        Models.Run.all.filter(
            **agent_filters,
            kind__in=[
                V1RunKind.JOB,
                V1RunKind.SERVICE,
                V1RunKind.TUNER,
                V1RunKind.NOTIFIER,
            ],
            live_state=LiveState.DELETION_PROGRESSING,
            status__in=LifeCycle.DONE_VALUES,
            pending__isnull=True,
            deleted_at__isnull=True,
            updated_at__lte=now().replace(second=0, microsecond=0)
            - timedelta(seconds=dj_settings.MIN_ARTIFACTS_DELETION_TIMEDELTA),
        )
        .prefetch_related("project")
        .values_list("uuid", "id", "pipeline_id")[:MAX_DELETE_ITEMS]
    )
    if deleting_runs:
        run_ids = [v[1] for v in deleting_runs]
        confirm_delete_runs(
            runs=Models.Run.all.filter(id__in=run_ids),
            run_ids=run_ids,
        )
        pipeline_ids = {v[2] for v in deleting_runs if v[2]}
        for pipeline_id in pipeline_ids:
            workers.send(
                SchedulerCeleryTasks.RUNS_CHECK_ORPHAN_PIPELINE,
                kwargs={"run_id": pipeline_id},
            )
    paths = [run[0].hex for run in deleting_runs]
    if paths:
        op = _operations.get_batch_cleaner_operation(
            connection=agent_config.artifacts_store,
            paths=paths,
            environment=V1DefaultScheduling.get_service_environment(
                agent_config.cleaner, agent_config.default_scheduling
            ),
            cleaner=agent_config.cleaner,
        )
        if is_new_agent:
            values.append(
                (
                    get_run_instance(owner_name, "agent", agent_id),
                    V1RunKind.JOB,
                    "cleaner",
                    op.to_json(include_version=True),
                    None,
                )
            )
        else:
            values.append(
                (
                    get_run_instance(owner_name, "agent", agent_id),
                    V1RunKind.JOB,
                    "cleaner",
                    op.to_json(include_version=True),
                )
            )

    # Clean and stop
    if dj_settings.HAS_ORG_MANAGEMENT:
        values_list = ["project__name", "uuid", "kind", "name", "id", "namespace"]
    else:
        values_list = ["project__name", "uuid", "kind", "name", "id"]
    deleting_runs = (
        Models.Run.all.filter(
            **agent_filters,
            kind__in={
                V1RunKind.JOB,
                V1RunKind.SERVICE,
                V1RunKind.TUNER,
                V1RunKind.NOTIFIER,
            },
            live_state=LiveState.DELETION_PROGRESSING,
            pending__isnull=True,
        )
        .exclude(status__in=LifeCycle.DONE_VALUES)
        .prefetch_related("project")
        .values_list(*values_list)[:max_budget]
    )
    if deleting_runs:
        Models.Run.all.filter(id__in=[v[4] for v in deleting_runs]).update(
            status=V1Statuses.STOPPED
        )
    for run in deleting_runs:
        run_uuid = run[1].hex
        if is_new_agent:
            values.append(
                (
                    get_run_instance(owner_name, run[0], run_uuid),
                    run[2],
                    run[3],
                    None,
                    run[5] if dj_settings.HAS_ORG_MANAGEMENT else None,
                )
            )
        else:
            values.append(
                (
                    get_run_instance(owner_name, run[0], run_uuid),
                    run[2],
                    run[3],
                    None,
                )
            )

    return paths, values, len(values) >= max_budget


def get_checks_runs(
    owner_name: str,
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    agent_filters: Optional[Dict] = None,
    is_new_agent: Optional[bool] = None,
) -> List[Tuple[str, str, str]]:
    agent_filters = agent_filters or {}
    start = now()
    end = start - timedelta(hours=1)
    if dj_settings.HAS_ORG_MANAGEMENT:
        values_list = ["project__name", "uuid", "kind", "id", "namespace"]
    else:
        values_list = ["project__name", "uuid", "kind", "id"]
    checks_runs = (
        Models.Run.objects.filter(
            **agent_filters,
            kind__in=[
                V1RunKind.JOB,
                V1RunKind.SERVICE,
                V1RunKind.TUNER,
                V1RunKind.NOTIFIER,
            ],
            checked_at__lte=end,
            managed_by=managed_by,
            status__in=LifeCycle.ON_K8S_VALUES,
        )
        .prefetch_related("project")
        .values_list(*values_list)
    )
    if is_new_agent:
        values = [
            (
                get_run_instance(owner_name, run[0], run[1].hex),
                run[2],
                run[4] if dj_settings.HAS_ORG_MANAGEMENT else None,
            )
            for run in checks_runs
        ]
    else:
        values = [
            (get_run_instance(owner_name, run[0], run[1].hex), run[2])
            for run in checks_runs
        ]
    if checks_runs:
        Models.Run.objects.filter(id__in=[r[3] for r in checks_runs]).update(
            checked_at=start
        )
    return values


def get_runs_by_pipeline(
    controller_id: int,
    pipeline_id: int,
    concurrency: int,
    consumed: int,
    max_budget: Optional[int],
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
) -> List[Models.Run]:
    num_to_run = get_num_to_start(
        concurrency=concurrency, consumed=consumed, max_budget=max_budget
    )
    if num_to_run < 1:
        return []

    return Models.Run.objects.filter(
        controller_id=controller_id,
        pipeline_id=pipeline_id,
        status=V1Statuses.COMPILED,
        pending__isnull=True,
        managed_by=managed_by,
    )[:num_to_run]


def check_pipelines(
    controller_id: int, max_budget: int
) -> Tuple[List[Models.Run], bool]:
    # TODO: TESTS
    # We queue runs managed by children pipelines
    to_update = []
    for pipeline in get_annotated_pipelines(controller_id=controller_id):
        runs = get_runs_by_pipeline(
            controller_id=controller_id,
            pipeline_id=pipeline[0],
            concurrency=pipeline[1],
            consumed=pipeline[2],
            max_budget=max_budget,
        )
        to_update += runs
        if max_budget:
            max_budget -= len(runs)
        if max_budget < 1:
            return to_update, True

    return to_update, False


def get_runs_by_controller(
    controller_id: int,
    concurrency: int,
    consumed: int,
    max_budget: Optional[int],
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
) -> List[Models.Run]:
    num_to_run = get_num_to_start(
        concurrency=concurrency, consumed=consumed, max_budget=max_budget
    )
    if num_to_run < 1:
        return []

    return Models.Run.objects.filter(
        controller_id=controller_id,
        pipeline_id=controller_id,
        status=V1Statuses.COMPILED,
        pending__isnull=True,
        managed_by=managed_by,
    )[:num_to_run]


def get_annotated_pipelines(
    controller_id: int, managed_by: Optional[ManagedBy] = ManagedBy.AGENT
) -> List[Tuple[int, int, int]]:
    pipelines = (
        Models.Run.objects.filter(
            kind__in=[V1RunKind.DAG, V1RunKind.MATRIX],
            status=V1Statuses.RUNNING,
            controller_id=controller_id,
            pipeline_runs__status=V1Statuses.COMPILED,
            pending__isnull=True,
            managed_by=managed_by,
        )
        .distinct()
        .values_list("id", flat=True)
    )
    pipelines = Models.Run.objects.filter(id__in=pipelines)
    pipelines = pipelines.annotate(
        consumed=Count(
            "pipeline_runs",
            filter=Q(
                pipeline_runs__status__in=LifeCycle.ON_K8S_VALUES | {V1Statuses.QUEUED}
            ),
            distinct=True,
        ),
    )
    return pipelines.values_list("id", "meta_info__concurrency", "consumed")


def get_annotated_controllers(
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    agent_filters: Optional[Dict] = None,
) -> List[Tuple[int, int, int]]:
    agent_filters = agent_filters or {}
    controllers = (
        Models.Run.objects.filter(
            **agent_filters,
            kind__in=[V1RunKind.DAG, V1RunKind.MATRIX],
            status=V1Statuses.RUNNING,
            pending__isnull=True,
            managed_by=managed_by,
            controller_id__isnull=True,
            controller_runs__status=V1Statuses.COMPILED,
        )
        .distinct()
        .values_list("id", flat=True)
    )
    controllers = Models.Run.objects.filter(id__in=controllers)
    controllers = controllers.annotate(
        consumed=Count(
            "pipeline_runs",
            filter=Q(
                pipeline_runs__status__in=LifeCycle.ON_K8S_VALUES | {V1Statuses.QUEUED}
            ),
            distinct=True,
        ),
    )
    return controllers.values_list("id", "meta_info__concurrency", "consumed")


def check_controllers(max_budget: int, agent_filters: Optional[Dict] = None) -> bool:
    if max_budget <= 0:
        return True

    full = False
    # TODO: TESTS
    to_update = []
    for controller in get_annotated_controllers(agent_filters=agent_filters):
        # We start by queueing directly managed runs by controller
        controller_budget = (
            controller[1]
            if (controller[1] is not None and controller[1] > 0)
            else max_budget
        )
        runs = get_runs_by_controller(
            controller_id=controller[0],
            concurrency=controller_budget,
            consumed=controller[2],
            max_budget=max_budget,
        )
        if controller_budget:
            controller_budget -= len(runs)
        to_update += runs

        # Check if controller still can queue
        if controller_budget < 1:
            full = True
            continue

        pipeline_runs, pipeline_full = check_pipelines(
            controller_id=controller[0], max_budget=controller_budget
        )
        to_update += pipeline_runs
        if pipeline_full:
            full = True

    # Split runs to pipeline and operations
    pipelines_to_update = [i for i in to_update if i.has_pipeline]
    runs_to_update = [i for i in to_update if not i.has_pipeline]
    # Start nested pipelines
    for pipeline in pipelines_to_update:
        workers.send(
            SchedulerCeleryTasks.RUNS_START_IMMEDIATELY,
            kwargs={"run_id": pipeline.id},
        )
    # Set queue
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.QUEUED,
        status="True",
        reason="AgentController",
        message="Run is queued",
    )
    bulk_new_run_status(runs_to_update, condition)
    return full


def get_queued_runs(
    managed_by: Optional[ManagedBy] = ManagedBy.AGENT,
    is_new_agent: Optional[bool] = None,
) -> Tuple[List[Tuple[str, str, str, str]], bool]:
    consumed = Models.Run.objects.filter(
        status__in=LifeCycle.ON_K8S_VALUES,
        pending__isnull=True,
        managed_by=managed_by,
    ).count()

    max_budget = dj_settings.MAX_CONCURRENCY - consumed

    if max_budget <= 0:
        return [], True

    num_to_run = get_num_to_start(
        concurrency=dj_settings.MAX_CONCURRENCY,
        consumed=consumed,
        max_budget=max_budget,
    )
    if num_to_run < 1:
        return [], True

    full = False
    if num_to_run > max_budget:
        full = True
        num_to_run = max_budget

    queryset = Models.Run.objects.filter(
        kind__in=[
            V1RunKind.JOB,
            V1RunKind.SERVICE,
            V1RunKind.TUNER,
            V1RunKind.NOTIFIER,
        ],
        status=V1Statuses.QUEUED,
        pending__isnull=True,
        managed_by=managed_by,
    ).prefetch_related("project")[:num_to_run]

    # Set scheduled
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.SCHEDULED,
        status="True",
        reason="AgentController",
        message="Operation is scheduled",
    )
    bulk_new_run_status(runs=[i for i in queryset], condition=condition)

    if is_new_agent:
        data = [
            (
                get_run_instance("default", run.project.name, run.uuid.hex),
                run.kind,
                run.name,
                run.content,
                getattr(run, "namespace", None),
            )
            for run in queryset
        ]
    else:
        data = [
            (
                get_run_instance("default", run.project.name, run.uuid.hex),
                run.kind,
                run.name,
                run.content,
            )
            for run in queryset
        ]

    return data, full


def get_agent_state() -> Dict:
    agent_config = settings.AGENT_CONFIG

    full = False
    # We check the schedules
    if check_schedules():
        full = True

    # We collect all jobs/services to stop
    stopping_runs, stopping_full = get_stopping_runs(
        owner_name="default",
        max_budget=dj_settings.MAX_CONCURRENCY,
        is_new_agent=True,
    )
    if stopping_full:
        full = True
    # We collect all jobs/services to delete
    if agent_config:
        _, deleting_runs, deleting_full = get_deleting_runs(
            owner_name="default",
            agent_id="agent",
            agent_config=agent_config,
            max_budget=dj_settings.MAX_CONCURRENCY,
            is_new_agent=True,
        )
        if deleting_full:
            full = True
    else:
        deleting_runs = []

    checks_runs = get_checks_runs(owner_name="default", is_new_agent=True)
    # We check the pipelines/controllers
    if check_controllers(max_budget=dj_settings.MAX_CONCURRENCY):
        full = True
    # We collect all jobs/services to start
    queued_runs, queued_full = get_queued_runs(is_new_agent=True)
    if queued_full:
        full = True
    return {
        V1Statuses.STOPPING: stopping_runs,
        V1Statuses.QUEUED: queued_runs,
        "checks": checks_runs,
        "deleting": deleting_runs,
        "full": full,
    }


def trigger_cron(state: bool = False) -> Dict:
    workers.send(CronsCeleryTasks.HEARTBEAT_OUT_OF_SYNC_SCHEDULES)
    workers.send(CronsCeleryTasks.HEARTBEAT_STOPPING_RUNS)
    workers.send(CronsCeleryTasks.HEARTBEAT_PROJECT_LAST_UPDATED)
    workers.send(CronsCeleryTasks.STATS_CALCULATION_PROJECTS)
    workers.send(CronsCeleryTasks.DELETE_ARCHIVED_PROJECTS)
    workers.send(CronsCeleryTasks.DELETE_IN_PROGRESS_PROJECTS)
    workers.send(CronsCeleryTasks.DELETE_ARCHIVED_RUNS)
    workers.send(CronsCeleryTasks.DELETE_IN_PROGRESS_RUNS)
    if not state:
        return {}

    # Return paths to be deleted not ops since the sandbox can manage the artifacts store
    deleting_paths, _, _ = get_deleting_runs(
        owner_name="default",
        agent_id="agent",
        agent_config=settings.AGENT_CONFIG,
        max_budget=dj_settings.MAX_CONCURRENCY,
    )
    return {
        "deleting": deleting_paths,
    }



from haupt.common.apis.regex import (
    OWNER_NAME_PATTERN,
    PROJECT_NAME_PATTERN,
    UUID_PATTERN,
)

# Projects
URLS_PROJECTS_CREATE = r"^{}/projects/create/?$".format(OWNER_NAME_PATTERN)
URLS_PROJECTS_LIST = r"^{}/projects/list/?$".format(OWNER_NAME_PATTERN)
URLS_PROJECTS_NAMES = r"^{}/projects/names/?$".format(OWNER_NAME_PATTERN)
URLS_PROJECTS_DETAILS = r"^{}/{}/?$".format(OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN)

# Resources
URLS_PROJECTS_RUNS_TAG = r"^{}/{}/runs/tag/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_STOP = r"^{}/{}/runs/stop/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_SKIP = r"^{}/{}/runs/skip/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_APPROVE = r"^{}/{}/runs/approve/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_DELETE = r"^{}/{}/runs/delete/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_SYNC = r"^{}/{}/runs/sync/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_ARTIFACTS_LINEAGE_V0 = r"^{}/{}/runs/artifacts_lineage/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_ARTIFACTS_LINEAGE = r"^{}/{}/runs/lineage/artifacts/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_LIST = r"^{}/{}/runs/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)

URLS_PROJECTS_ARCHIVE = r"^{}/{}/archive/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RESTORE = r"^{}/{}/restore/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_BOOKMARK = r"^{}/{}/bookmark/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_UNBOOKMARK = r"^{}/{}/unbookmark/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)

URLS_PROJECTS_RUNS_INVALIDATE = r"^{}/{}/runs/invalidate/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_BOOKMARK = r"^{}/{}/runs/bookmark/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_ACTIVITIES = r"^{}/{}/activities/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)

URLS_PROJECTS_ARTIFACT_UPLOAD = r"^{}/{}/artifacts/{}/upload?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN, UUID_PATTERN
)
URLS_PROJECTS_RUNS_TRANSFER = r"^{}/{}/runs/transfer/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_ARCHIVE = r"^{}/{}/runs/archive/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_RUNS_RESTORE = r"^{}/{}/runs/restore/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)
URLS_PROJECTS_STATS = r"^{}/{}/stats/?$".format(
    OWNER_NAME_PATTERN, PROJECT_NAME_PATTERN
)



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


def read_readme():
    if not os.path.exists("./README.md"):
        return ""
    with open("./README.md") as f:
        return f.read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


with open(os.path.join("./haupt/pkg.py"), encoding="utf8") as f:
    pkg = {}
    exec(f.read(), pkg)


with open("requirements/requirements.txt") as requirements_file:
    requirements = requirements_file.read().splitlines()

with open("requirements/platform.txt") as requirements_file:
    platform_requirements = requirements_file.read().splitlines()

with open("requirements/sandbox.txt") as requirements_file:
    sandbox_requirements = requirements_file.read().splitlines()

with open("requirements/dev.txt") as requirements_file:
    dev_requirements = requirements_file.read().splitlines()

with open("requirements/fs.txt") as requirements_file:
    fs_requirements = requirements_file.read().splitlines()

with open("requirements/streams.txt") as requirements_file:
    stream_requirements = requirements_file.read().splitlines()

if os.environ.get("USE_PROD_PACKAGES"):
    with open("requirements/prod.txt") as requirements_file:
        requirements += requirements_file.read().splitlines()

extra = {
    "platform": platform_requirements,
    "sandbox": sandbox_requirements + stream_requirements,
    "fs": fs_requirements,
    "streams": stream_requirements,
    "dev": dev_requirements,
}

setup(
    name=pkg["NAME"],
    version=pkg["VERSION"],
    description=pkg["DESC"],
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    maintainer=pkg["AUTHOR"],
    maintainer_email=pkg["EMAIL"],
    author=pkg["AUTHOR"],
    author_email=pkg["EMAIL"],
    url=pkg["URL"],
    license=pkg["LICENSE"],
    platforms="any",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    keywords=[
        "polyaxon",
        "aws",
        "s3",
        "microsoft",
        "azure",
        "google cloud storage",
        "gcs",
        "deep-learning",
        "machine-learning",
        "data-science",
        "neural-networks",
        "artificial-intelligence",
        "ai",
        "reinforcement-learning",
        "kubernetes",
        "aws",
        "microsoft",
        "azure",
        "google cloud",
        "tensorFlow",
        "pytorch",
        "paddlepaddle",
        "matplotlib",
        "plotly",
        "visualization",
        "analytics",
        "hyperopt",
        "scikit-learn",
        "optimization",
        "bayesian-optimization",
    ],
    install_requires=[requirements],
    extras_require=extra,
    entry_points={"console_scripts": ["haupt = haupt.main:cli"]},
    python_requires=">=3.8",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Typing :: Typed",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    tests_require=["pytest"],
    cmdclass={"test": PyTest},
)



from haupt.common import auditor
from haupt.common.events.registry import project

auditor.subscribe(project.ProjectCreatedActorEvent)
auditor.subscribe(project.ProjectUpdatedActorEvent)
auditor.subscribe(project.ProjectViewedActorEvent)
auditor.subscribe(project.ProjectSettingsActorEvent)
auditor.subscribe(project.ProjectDeletedActorEvent)
auditor.subscribe(project.ProjectStatsActorEvent)
auditor.subscribe(project.ProjectRunsViewedActorEvent)



FEATURES = "FEATURES"
SCHEDULER = "SCHEDULER"
STATS = "STATS"
TTL = "TTL"
SIDECARS = "SIDECARS"
INIT = "INIT"
CLEANING_INTERVALS = "CLEANING_INTERVALS"
COMPONENT = "COMPONENT"
TIMEOUT = "TIMEOUT"
EMAIL = "EMAIL"
USER = "USER"



import os

from typing import Optional

from polyaxon._connections import CONNECTION_CONFIG
from polyaxon._contexts import paths as ctx_paths
from polyaxon._env_vars.getters import get_artifacts_store_name
from polyaxon._env_vars.keys import ENV_KEYS_SERVICE_MODE
from polyaxon._fs.fs import close_fs, get_async_fs_from_connection, get_default_fs
from polyaxon._fs.types import FSSystem
from polyaxon._services import PolyaxonServices


class AppFS:
    _connections = {}

    @classmethod
    async def set_fs(cls, connection: Optional[str] = None) -> FSSystem:
        fs = None
        if connection:
            name = connection
            connection = CONNECTION_CONFIG.get_connection_for(connection)
            if connection:
                fs = await get_async_fs_from_connection(connection=connection)
        else:
            fs = await get_default_fs()
            name = get_artifacts_store_name()

        cls._connections[name] = fs
        return cls._connections[name]

    @classmethod
    async def close_fs(cls, connection: Optional[str] = None):
        connection = connection or get_artifacts_store_name()
        fs = cls._connections.get(connection)
        await close_fs(fs)

    @classmethod
    async def get_fs(cls, connection: Optional[str] = None) -> FSSystem:
        if (
            not connection
            or os.environ.get(ENV_KEYS_SERVICE_MODE) == PolyaxonServices.VIEWER
        ):
            connection = get_artifacts_store_name()
        fs = cls._connections.get(connection)
        if not fs:
            return await cls.set_fs(connection=connection)
        return fs

    @staticmethod
    def get_fs_root_path(connection: Optional[str] = None) -> str:
        connection = connection or get_artifacts_store_name()
        connection = CONNECTION_CONFIG.get_connection_for(connection)
        if not connection:
            return ctx_paths.CONTEXT_ARTIFACTS_ROOT
        return connection.store_path



from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.conf import settings

from haupt.apis.serializers.base.tags import TagsMixin
from haupt.common import auditor
from haupt.common.authentication.base import is_normal_user
from haupt.common.content_types import ContentTypes
from haupt.common.events.registry.archive import RUN_ARCHIVED_ACTOR, RUN_RESTORED_ACTOR
from haupt.common.events.registry.run import (
    RUN_APPROVED_ACTOR,
    RUN_DELETED_ACTOR,
    RUN_DONE,
    RUN_SKIPPED_ACTOR,
    RUN_STOPPED_ACTOR,
    RUN_TRANSFERRED_ACTOR,
)
from haupt.db.defs import Models
from haupt.db.managers.bookmarks import bookmark_obj
from haupt.db.managers.statuses import bulk_new_run_status
from polyaxon.schemas import (
    LifeCycle,
    LiveState,
    V1RunPending,
    V1StatusCondition,
    V1Statuses,
)


def create_runs_tags(view, request, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    tags = request.data.get("tags", [])
    if not tags:
        return Response(status=status.HTTP_200_OK, data={})

    updated = []
    queryset = view.enrich_queryset(Models.Run.all)
    queryset = queryset.filter(uuid__in=uuids)
    for run in queryset.only("id", "tags"):
        run.tags = TagsMixin.validated_tags({"tags": tags, "merge": True}, run.tags)[
            "tags"
        ]
        updated.append(run)

        Models.Run.objects.bulk_update(updated, ["tags"])
    return Response(status=status.HTTP_200_OK, data={})


def stop_runs(view, request, actor, *args, **kwargs):
    # For Audit
    view.set_owner()

    uuids = request.data.get("uuids", [])
    # Immediate stop
    queryset = view.enrich_queryset(Models.Run.all)
    queryset = queryset.filter(uuid__in=uuids)
    queryset = queryset.filter(status__in=LifeCycle.SAFE_STOP_VALUES)
    runs = [r for r in queryset]
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.STOPPED,
        status="True",
        reason="EventHandler",
        message="User requested to stop the run.",
    )
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(actor):
        status_meta_info = {
            "user": {"username": actor.username, "email": actor.email},
        }
        condition.meta_info = status_meta_info
    bulk_new_run_status(queryset, condition)
    # For Audit
    for run in runs:
        auditor.record(
            event_type=RUN_STOPPED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )
        auditor.record(event_type=RUN_DONE, instance=run, previous_status=run.status)

    queryset = view.enrich_queryset(Models.Run.all)
    queryset = queryset.filter(uuid__in=uuids)
    queryset = queryset.exclude(status__in=LifeCycle.DONE_OR_IN_PROGRESS_VALUES)
    runs = [r for r in queryset]
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.STOPPING,
        status="True",
        reason="EventHandler",
        message="User requested to stop the run.",
    )
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(actor):
        status_meta_info = {
            "user": {"username": actor.username, "email": actor.email},
        }
        condition.meta_info = status_meta_info
    bulk_new_run_status(runs, condition)
    # For Audit
    for run in runs:
        auditor.record(
            event_type=RUN_STOPPED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )
        auditor.record(event_type=RUN_DONE, instance=run, previous_status=run.status)

    return Response(status=status.HTTP_200_OK, data={})


def skip_runs(view, request, actor, *args, **kwargs):
    # For Audit
    view.set_owner()

    uuids = request.data.get("uuids", [])
    # Immediate stop
    queryset = view.enrich_queryset(Models.Run.restorable)
    queryset = queryset.filter(uuid__in=uuids)
    queryset = queryset.filter(status__in=LifeCycle.SAFE_STOP_VALUES)
    runs = [r for r in queryset]
    condition = V1StatusCondition.get_condition(
        type=V1Statuses.SKIPPED,
        status="True",
        reason="EventHandler",
        message="User requested to skip the run.",
    )
    if settings.HAS_ORG_MANAGEMENT and is_normal_user(actor):
        status_meta_info = {
            "user": {"username": actor.username, "email": actor.email},
        }
        condition.meta_info = status_meta_info
    bulk_new_run_status(queryset, condition)
    # For Audit
    for run in runs:
        auditor.record(
            event_type=RUN_SKIPPED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )
        auditor.record(event_type=RUN_DONE, instance=run, previous_status=run.status)

    return Response(status=status.HTTP_200_OK, data={})


def approve_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    queryset = view.enrich_queryset(Models.Run.objects)
    queryset = queryset.filter(uuid__in=uuids)
    queryset = queryset.filter(
        pending__in={V1RunPending.APPROVAL, V1RunPending.CACHE},
    )
    runs = [r for r in queryset]
    queryset.update(pending=None)
    # For Audit
    view.set_owner()
    for run in runs:
        auditor.record(
            event_type=RUN_APPROVED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )

    return Response(status=status.HTTP_200_OK, data={})


def delete_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    queryset = view.enrich_queryset(Models.Run.restorable)
    runs = queryset.filter(uuid__in=uuids)
    # For Audit
    view.set_owner()
    for run in runs:
        auditor.record(
            event_type=RUN_DELETED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )
    # Deletion in progress
    runs.update(live_state=LiveState.DELETION_PROGRESSING)
    return Response(status=status.HTTP_200_OK, data={})


def invalidate_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    queryset = view.enrich_queryset(Models.Run.all)
    queryset.filter(uuid__in=uuids).update(state=None)
    return Response(status=status.HTTP_200_OK, data={})


def archive_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    queryset = view.enrich_queryset(Models.Run.objects)
    queryset = queryset.filter(uuid__in=uuids)
    # Set to stopping all runs that are not ended yet
    queryset.exclude(status__in=LifeCycle.DONE_OR_IN_PROGRESS_VALUES).update(
        status=V1Statuses.STOPPING,
    )
    # Pipeline runs
    Models.Run.objects.filter(pipeline__uuid__in=uuids).update(
        live_state=LiveState.ARCHIVED
    )
    runs = [r for r in queryset]
    queryset.update(live_state=LiveState.ARCHIVED)
    # For Audit
    view.set_owner()
    for run in runs:
        auditor.record(
            event_type=RUN_ARCHIVED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )

    return Response(status=status.HTTP_200_OK, data={})


def restore_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    queryset = view.enrich_queryset(Models.Run.archived)
    queryset = queryset.filter(uuid__in=uuids)
    # Restore pipeline runs
    Models.Run.archived.filter(pipeline__uuid__in=uuids).update(
        live_state=LiveState.LIVE
    )
    runs = [r for r in queryset]
    queryset.update(live_state=LiveState.LIVE)
    # For Audit
    view.set_owner()
    for run in runs:
        auditor.record(
            event_type=RUN_RESTORED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )

    return Response(status=status.HTTP_200_OK, data={})


def bookmarks_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    user = request.user if settings.HAS_ORG_MANAGEMENT else None
    content_type = ContentTypes.RUN.value
    queryset = view.enrich_queryset(Models.Run.all)
    for run in queryset.filter(uuid__in=uuids).only("id"):
        bookmark_obj(user=user, obj=run, content_type=content_type)

    return Response(status=status.HTTP_200_OK, data={})


def transfer_runs(view, request, actor, *args, **kwargs):
    uuids = request.data.get("uuids", [])
    project_name = request.data.get("project")
    if project_name == view.project_name:
        return Response(status=status.HTTP_200_OK, data={})
    if not project_name:
        raise ValidationError("The destination project was not provided.")
    try:
        filters = {"name": project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["owner__name"] = view.owner_name
        dest_project = Models.Project.objects.only("id").get(**filters)
    except Models.Project.DoesNotExist:
        raise ValidationError(
            "The destination project `{}` does not exist.".format(project_name)
        )

    queryset = view.enrich_queryset(Models.Run.all)
    queryset = queryset.filter(uuid__in=uuids)
    run_ids = [i for i in queryset.values_list("id", flat=True)]
    queryset.update(project_id=dest_project.id)
    # For Audit
    view.set_owner()
    for run in Models.Run.all.filter(id__in=run_ids):
        auditor.record(
            event_type=RUN_TRANSFERRED_ACTOR,
            instance=run,
            actor_id=actor.id,
            actor_name=actor.username,
            owner_id=view._owner_id,
            owner_name=view.owner_name,
            project_name=view.project_name,
        )
    return Response(status=status.HTTP_200_OK, data={})



import click


@click.command()
@click.option(
    "--host",
    help="The service host.",
)
@click.option(
    "--port",
    type=int,
    help="The service port.",
)
@click.option(
    "--workers",
    type=int,
    help="Number of workers.",
)
@click.option(
    "--per-core",
    is_flag=True,
    default=False,
    help="To enable workers per core.",
)
@click.option(
    "--uds",
    help="UNIX domain socket binding.",
)
def streams(host: str, port: int, workers: int, per_core: bool, uds: str):
    """Start a new streams session."""
    from haupt.cli.runners.streams import start

    return start(host=host, port=port, workers=workers, per_core=per_core, uds=uds)



from clipped.utils.versions import compare_versions
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.conf import settings
from django.http import Http404

from haupt.apis.endpoints.project import VersionEndpoint, VersionListEndpoint
from haupt.apis.methods import entity_stages as methods
from haupt.apis.serializers.project_versions import (
    ArtifactVersionDetailSerializer,
    ComponentVersionDetailSerializer,
    ModelVersionDetailSerializer,
    ProjectVersionDetailSerializer,
    ProjectVersionNameSerializer,
    ProjectVersionSerializer,
    ProjectVersionStageSerializer,
)
from haupt.common import auditor
from haupt.common.apis.filters import OrderingFilter, QueryFilter
from haupt.common.apis.paginator import LargeLimitOffsetPagination
from haupt.common.endpoints.base import (
    CreateEndpoint,
    DestroyEndpoint,
    ListEndpoint,
    RetrieveEndpoint,
    UpdateEndpoint,
)
from haupt.common.events.registry.artifact_version import (
    ARTIFACT_VERSION_CREATED_ACTOR,
    ARTIFACT_VERSION_DELETED_ACTOR,
    ARTIFACT_VERSION_NEW_STAGE,
    ARTIFACT_VERSION_TRANSFERRED_ACTOR,
    ARTIFACT_VERSION_UPDATED_ACTOR,
    ARTIFACT_VERSION_VIEWED_ACTOR,
)
from haupt.common.events.registry.component_version import (
    COMPONENT_VERSION_CREATED_ACTOR,
    COMPONENT_VERSION_DELETED_ACTOR,
    COMPONENT_VERSION_NEW_STAGE,
    COMPONENT_VERSION_TRANSFERRED_ACTOR,
    COMPONENT_VERSION_UPDATED_ACTOR,
    COMPONENT_VERSION_VIEWED_ACTOR,
)
from haupt.common.events.registry.model_version import (
    MODEL_VERSION_CREATED_ACTOR,
    MODEL_VERSION_DELETED_ACTOR,
    MODEL_VERSION_NEW_STAGE,
    MODEL_VERSION_TRANSFERRED_ACTOR,
    MODEL_VERSION_UPDATED_ACTOR,
    MODEL_VERSION_VIEWED_ACTOR,
)
from haupt.db.defs import Models
from haupt.db.managers.versions import add_version_contributors
from haupt.db.query_managers.project_version import ProjectVersionQueryManager
from polyaxon.schemas import V1ProjectVersionKind

ADDITIONAL_SELECT_RELATED = ["project__owner"] if settings.HAS_ORG_MANAGEMENT else []


class ProjectVersionListView(VersionListEndpoint, ListEndpoint, CreateEndpoint):
    queryset = Models.ProjectVersion.objects.defer("content").order_by("-updated_at")
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = ProjectVersionQueryManager
    check_alive = ProjectVersionQueryManager.CHECK_ALIVE
    ordering = ProjectVersionQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ProjectVersionQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ProjectVersionQueryManager.FIELDS_ORDERING_PROXY
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True
    ALLOWED_METHODS = ["GET", "POST"]
    EVENTS_TYPE = None

    def perform_create(self, serializer):
        instance = serializer.save(project=self.project)
        if not self.EVENTS_TYPE:
            return
        add_version_contributors(instance, users=[self.request.user])
        auditor.record(
            event_type=self.EVENTS_TYPE,
            instance=instance,
            actor_id=self.request.user.id,
            actor_name=self.request.user.username,
            owner_id=self.project.owner_id,
            owner_name=self.owner_name,
            hub_name=self.project_name,
        )


class ProjectComponentVersionListView(ProjectVersionListView):
    serializer_class_mapping = {
        "GET": ProjectVersionSerializer,
        "POST": ComponentVersionDetailSerializer,
    }
    EVENTS_TYPE = COMPONENT_VERSION_CREATED_ACTOR
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.COMPONENT)
        .defer("content")
        .order_by("-updated_at")
    )


class ProjectModelVersionListView(ProjectVersionListView):
    serializer_class_mapping = {
        "GET": ProjectVersionSerializer,
        "POST": ModelVersionDetailSerializer,
    }
    EVENTS_TYPE = MODEL_VERSION_CREATED_ACTOR
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.MODEL)
        .defer("content")
        .order_by("-updated_at")
    )


class ProjectArtifactVersionListView(ProjectVersionListView):
    serializer_class_mapping = {
        "GET": ProjectVersionSerializer,
        "POST": ArtifactVersionDetailSerializer,
    }
    EVENTS_TYPE = ARTIFACT_VERSION_CREATED_ACTOR
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.ARTIFACT)
        .defer("content")
        .order_by("-updated_at")
    )


class ProjectVersionNameListView(VersionListEndpoint, ListEndpoint):
    """List project versions' names for a user."""

    queryset = Models.ProjectVersion.objects.only("name").order_by("-updated_at")
    filter_backends = (QueryFilter, OrderingFilter)
    query_manager = ProjectVersionQueryManager
    check_alive = ProjectVersionQueryManager.CHECK_ALIVE
    ordering = ProjectVersionQueryManager.FIELDS_DEFAULT_ORDERING
    ordering_fields = ProjectVersionQueryManager.FIELDS_ORDERING
    ordering_proxy_fields = ProjectVersionQueryManager.FIELDS_ORDERING_PROXY
    serializer_class = ProjectVersionNameSerializer
    pagination_class = LargeLimitOffsetPagination
    ALLOWED_METHODS = ["GET"]


class ProjectComponentVersionNameListView(ProjectVersionNameListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.COMPONENT)
        .only("name")
        .order_by("-updated_at")
    )


class ProjectModelVersionNameListView(ProjectVersionNameListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.MODEL)
        .only("name")
        .order_by("-updated_at")
    )


class ProjectArtifactVersionNameListView(ProjectVersionNameListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.ARTIFACT)
        .only("name")
        .order_by("-updated_at")
    )


ADDITIONAL_SELECT_RELATED_DETAILS = ADDITIONAL_SELECT_RELATED + (
    [
        "connection",
        "connection__agent",
    ]
    if settings.HAS_ORG_MANAGEMENT
    else []
)
ADDITIONAL_PREFETCH_RELATED_DETAILS = (
    [
        "contributors",
        "lineage__connection",
    ]
    if settings.HAS_ORG_MANAGEMENT
    else []
)


class ProjectVersionDetailView(
    VersionEndpoint, RetrieveEndpoint, UpdateEndpoint, DestroyEndpoint
):
    """
    get:
        Get a project version details.
    patch:
        Update a project version details.
    delete:
        Delete a project version.
    """

    queryset = Models.ProjectVersion.objects.select_related(
        *ADDITIONAL_SELECT_RELATED_DETAILS,
        "project",
        "run",
        "run__project",
    ).prefetch_related(
        *ADDITIONAL_PREFETCH_RELATED_DETAILS,
        "lineage",
        "lineage__artifact",
    )
    ALLOWED_METHODS = ["GET", "PUT", "PATCH", "DELETE"]
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True

    def get_serializer(self, *args, **kwargs):
        service = self.get_service(request=self.request)
        service_version = self.get_service_version(request=self.request)
        if not service:
            compatibility = "both"
        elif service_version and compare_versions(
            current=service_version, reference="1.18.0", comparator="<"
        ):
            compatibility = "old"
        else:
            compatibility = "new"

        serializer_class = self.get_serializer_class()
        context = self.get_serializer_context()
        context["compatibility"] = compatibility
        kwargs["context"] = context
        return serializer_class(*args, **kwargs)

    def perform_update(self, serializer):
        instance = serializer.save()
        add_version_contributors(instance, users=[self.request.user])


class ProjectComponentVersionDetailView(ProjectVersionDetailView):
    serializer_class_mapping = {
        "GET": ProjectVersionDetailSerializer,
        "DELETE": ProjectVersionDetailSerializer,
        "PATCH": ComponentVersionDetailSerializer,
        "PUT": ComponentVersionDetailSerializer,
    }
    queryset = ProjectVersionDetailView.queryset.filter(
        kind=V1ProjectVersionKind.COMPONENT
    )
    AUDITOR_EVENT_TYPES = {
        "GET": COMPONENT_VERSION_VIEWED_ACTOR,
        "PUT": COMPONENT_VERSION_UPDATED_ACTOR,
        "PATCH": COMPONENT_VERSION_UPDATED_ACTOR,
        "DELETE": COMPONENT_VERSION_DELETED_ACTOR,
    }


class ProjectModelVersionDetailView(ProjectVersionDetailView):
    serializer_class_mapping = {
        "GET": ProjectVersionDetailSerializer,
        "DELETE": ProjectVersionDetailSerializer,
        "PATCH": ModelVersionDetailSerializer,
        "PUT": ModelVersionDetailSerializer,
    }
    queryset = ProjectVersionDetailView.queryset.filter(kind=V1ProjectVersionKind.MODEL)
    AUDITOR_EVENT_TYPES = {
        "GET": MODEL_VERSION_VIEWED_ACTOR,
        "PUT": MODEL_VERSION_UPDATED_ACTOR,
        "PATCH": MODEL_VERSION_UPDATED_ACTOR,
        "DELETE": MODEL_VERSION_DELETED_ACTOR,
    }


class ProjectArtifactVersionDetailView(ProjectVersionDetailView):
    serializer_class_mapping = {
        "GET": ProjectVersionDetailSerializer,
        "DELETE": ProjectVersionDetailSerializer,
        "PATCH": ArtifactVersionDetailSerializer,
        "PUT": ArtifactVersionDetailSerializer,
    }
    queryset = ProjectVersionDetailView.queryset.filter(
        kind=V1ProjectVersionKind.ARTIFACT
    )
    AUDITOR_EVENT_TYPES = {
        "GET": ARTIFACT_VERSION_VIEWED_ACTOR,
        "PUT": ARTIFACT_VERSION_UPDATED_ACTOR,
        "PATCH": ARTIFACT_VERSION_UPDATED_ACTOR,
        "DELETE": ARTIFACT_VERSION_DELETED_ACTOR,
    }


class ProjectVersionTransferView(VersionEndpoint, CreateEndpoint):
    queryset = Models.ProjectVersion.objects.defer(
        "content", "description"
    ).select_related(
        *ADDITIONAL_SELECT_RELATED,
        "project",
    )
    ALLOWED_METHODS = ["POST"]
    AUDIT_OWNER = True
    AUDIT_PROJECT = True
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        project_name = request.data.get("project")
        if project_name == self.project_name:
            return Response(status=status.HTTP_200_OK, data={})
        if not project_name:
            raise ValidationError("The destination project was not provided.")
        try:
            org_filters = (
                dict(owner__name=self.owner_name) if settings.HAS_ORG_MANAGEMENT else {}
            )
            dest_project = Models.Project.objects.only("id").get(
                **org_filters, name=project_name
            )
        except Models.Project.DoesNotExist:
            raise ValidationError(
                "The destination project `{}` does not exist.".format(project_name)
            )

        self.version.project_id = dest_project.id
        self.version.save(update_fields=["project_id", "updated_at"])
        add_version_contributors(self.version, users=[request.user])
        self.audit(request, *args, **kwargs)
        return Response(status=status.HTTP_200_OK, data={})


class ProjectComponentVersionTransferView(ProjectVersionTransferView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.COMPONENT)
        .defer("content", "description")
        .select_related(*ADDITIONAL_SELECT_RELATED, "project")
    )
    AUDITOR_EVENT_TYPES = {"POST": COMPONENT_VERSION_TRANSFERRED_ACTOR}


class ProjectModelVersionTransferView(ProjectVersionTransferView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.MODEL)
        .defer("content", "description")
        .select_related(
            *ADDITIONAL_SELECT_RELATED,
            "project",
        )
    )
    AUDITOR_EVENT_TYPES = {"POST": MODEL_VERSION_TRANSFERRED_ACTOR}


class ProjectArtifactVersionTransferView(ProjectVersionTransferView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.ARTIFACT)
        .defer("content", "description")
        .select_related(
            *ADDITIONAL_SELECT_RELATED,
            "project",
        )
    )
    AUDITOR_EVENT_TYPES = {"POST": ARTIFACT_VERSION_TRANSFERRED_ACTOR}


class ProjectVersionStageListView(VersionEndpoint, RetrieveEndpoint, CreateEndpoint):
    queryset = Models.ProjectVersion.objects.defer(
        "content", "description"
    ).select_related(
        *ADDITIONAL_SELECT_RELATED,
        "project",
    )
    serializer_class = ProjectVersionStageSerializer
    ALLOWED_METHODS = ["POST", "GET"]

    def perform_create(self, serializer):
        try:
            methods.create_stage(
                view=self,
                serializer=serializer,
                event_type=self.get_event_type(self.request),
            )
        except Models.ProjectVersion.DoesNotExist:
            raise Http404


class ProjectComponentVersionStageListView(ProjectVersionStageListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.COMPONENT)
        .defer("content", "description")
        .select_related(
            *ADDITIONAL_SELECT_RELATED,
            "project",
        )
    )
    AUDITOR_EVENT_TYPES = {"POST": COMPONENT_VERSION_NEW_STAGE}


class ProjectModelVersionStageListView(ProjectVersionStageListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.MODEL)
        .defer("content", "description")
        .select_related(
            *ADDITIONAL_SELECT_RELATED,
            "project",
        )
    )
    AUDITOR_EVENT_TYPES = {"POST": MODEL_VERSION_NEW_STAGE}


class ProjectArtifactVersionStageListView(ProjectVersionStageListView):
    queryset = (
        Models.ProjectVersion.objects.filter(kind=V1ProjectVersionKind.ARTIFACT)
        .defer("content", "description")
        .select_related(
            *ADDITIONAL_SELECT_RELATED,
            "project",
        )
    )
    AUDITOR_EVENT_TYPES = {"POST": ARTIFACT_VERSION_NEW_STAGE}



from haupt.db.query_managers.manager import BaseQueryManager
from polyaxon._pql.builder import (
    ArrayCondition,
    DateTimeCondition,
    SearchCondition,
    ValueCondition,
)
from polyaxon._pql.parser import (
    parse_datetime_operation,
    parse_search_operation,
    parse_value_operation,
)


class ProjectVersionQueryManager(BaseQueryManager):
    NAME = "project_version"
    FIELDS_USE_NAME = {
        "project",
    }
    FIELDS_PROXY = {
        "id": "uuid",
        "uid": "uuid",
        "teams": "project__teams__name",
    }
    FIELDS_ORDERING = (
        "created_at",
        "updated_at",
        "name",
        "uuid",
        "state",
        "stage",
        "kind",
    )
    FIELDS_USE_UUID = {
        "run",
    }
    CHECK_ALIVE = False
    PARSERS_BY_FIELD = {
        # Uuid
        "id": parse_search_operation,
        "uid": parse_search_operation,
        "uuid": parse_search_operation,
        # Dates
        "created_at": parse_datetime_operation,
        "updated_at": parse_datetime_operation,
        # Name
        "name": parse_search_operation,
        # Description
        "description": parse_search_operation,
        # Tags
        "tags": parse_value_operation,
        # Kind
        "kind": parse_value_operation,
        # State
        "state": parse_value_operation,
        # Stage
        "stage": parse_value_operation,
        # Run
        "run": parse_value_operation,
        # Teams
        "teams": parse_value_operation,
        # Projects
        "project": parse_value_operation,
    }
    CONDITIONS_BY_FIELD = {
        # Uuid
        "id": SearchCondition,
        "uid": SearchCondition,
        "uuid": SearchCondition,
        # Dates
        "created_at": DateTimeCondition,
        "updated_at": DateTimeCondition,
        # Name
        "name": SearchCondition,
        # Description
        "description": SearchCondition,
        # Tags
        "tags": ArrayCondition,
        # Kind
        "kind": ValueCondition,
        # State
        "state": ValueCondition,
        # Stage
        "stage": ValueCondition,
        # Run
        "run": ValueCondition,
        # Teams
        "teams": ValueCondition,
        # Projects
        "project": ValueCondition,
    }
